!-----------------------------------------------------------------------
!       purpose
!       -------
!       perform all calculations needed for TM5 test run  
!       with 1 ppb concentration
!
!       interface
!       ---------
!       call trace0
!       call trace1
!       call source
!       call chemie
!
!       method
!       ------
!       subroutine trace0 is called from start, and is used 
!                to input tracer specific values.
!       subroutine trace1 is called from start when istart = 2, and provides
!                a way to do some initialization, that has to be done only
!                at the beginning of a run.
!       subroutine source is called every nsrce seconds and provides 
!                the source/sink calculation.
!       subroutine chem  is called every nchem seconds and provides the
!                chemistry calculation on the tracer fields.
!       in zoom mode, recursive calling of the children is performed
!       
!-----------------------------------------------------------------------

module sources_sinks

  use GO         , only : gol, goErr, goPr
  use dims       , only : nregions
  use tracer_data, only : chem_dat

  implicit none

  ! --- in/out --------------------------
  
  private
  
  public  ::  Sources_sinks_Init, Sources_sinks_Done
  public  ::  Trace0, Trace1, Trace_End, trace_after_read
  public  ::  Source1, Source2
  public  ::  chem_dat
  
  ! --- const --------------------------------------

  character(len=*), parameter  ::  mname = 'sources_sinks'


contains


  ! ================================================================
  
  
  subroutine Sources_Sinks_Init( status )
  
    ! --- in/out --------------------------------
    
    integer, intent(out)           ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Sources_Sinks_Init'
    
    ! --- begin --------------------------------
    
    ! nothing to be init

    ! ok
    status = 0
    
  end subroutine Sources_Sinks_Init
  
  
  ! ***
  
  
  subroutine Sources_Sinks_Done( status )
  
    ! --- in/out --------------------------------
    
    integer, intent(out)           ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Sources_Sinks_Done'

    ! --- begin --------------------------------
    
    ! nothing to be done

    ! ok
    status = 0
    
  end subroutine Sources_Sinks_Done
  
  
  
  ! ================================================================
  

  !-----------------------------------------------------------------------
  !      
  !       trace0
  !
  !
  !       purpose
  !       -------
  !       initialise data needed for chemistry, 
  !       e.g. read emissions and boundary conditions
  !       routine is calculated at start and at beginning each month...
  !
  !       interface
  !       ---------
  !       call trace0
  !
  !       method
  !       ------
  !       subroutine trace0 is called from module input, and provides a way
  !                to input tracer specific values.
  !       
  !       external
  !       --------
  !           calc_sm
  !           ini_zone
  !           ini_zoneg
  !           call_diagbud
  !           call_diagbudg
  !  
  !       reference
  !       ---------
  !       see above
  !---------------------------------------------------------------------   

  subroutine trace0( status )

    use dims,        only : im, jm, lm
    use dims,          only : newmonth, idate, mlen, dxy11, nlat180, newsrun
    use dims,          only : sec_day,sec_month,sec_year
    use geometry,      only : calc_dxy
    use datetime,      only : calc_sm
#ifdef with_budgets
    use budget_fine,   only : diagbud,  ini_zone
    use budget_global, only : diagbudg, ini_zoneg
#endif
    use ParTools     , only : lmloc

    ! --- in/out --------------------------

    integer, intent(out)        ::  status

    ! --- local --------------------------

    integer  :: xref_, yref_, zref_, tref_
    integer  :: region
    integer  :: i, j, ii, n, nmonth
    integer  :: io, sfstart, ifail, sfend
    real     :: year2s, month2s, year2month
    integer  :: iyear, nxw, nyw, nzw, nmw

    ! --- begin ---------------------------

    write (gol,'("trace0: newmonth ",l1)') newmonth; call goPr
    write (gol,'("trace0: start month ",i2)') idate(2); call goPr

    ! calculate some conversion factors related to time...

    call calc_sm(mlen,sec_day,sec_month,sec_year)

    ! year2month: scale factor for yearly total to specific month
    year2month = sec_month/sec_year
    month2s=1.0/sec_month            ! conversion to emission per timestep
    year2s =1.0/sec_year  
    iyear  = idate(1)

    !print *,'trace0: seconds in a day, month and year', &
    !     sec_day,sec_month,sec_year,year2month
    
    !
    ! calculate the area of the gridboxes on 1x1 degree
    !
    call calc_dxy(dxy11,nlat180)

#ifdef with_budgets
    call ini_zone    ! the budget of region = nregions...
    call ini_zoneg   ! the global budget

    do n=1,nregions
       call diagbud (n,1)               ! initialize budget files fine
       call diagbudg(n,1)               ! initialize budget files global
    end do
#endif

    write (gol,'("trace0: end month ",i2)') idate(2); call goPr

    ! ok
    status = 0

  end subroutine trace0



  subroutine trace1
    !----------------------------------------------------------------------- 
    !
    !       purpose
    !       -------
    !       this subroutine determines the initial values for tracer mass and
    !       and its slopes.  it is called from input when istart = 2.
    !
    !       interface
    !       ---------
    !       call trace0
    !
    !       method
    !       ------
    !         
    !       reference
    !       ---------
    !       see above
    !
    !----------------------------------------------------------------------- 

    use dims,        only : im, jm, lm, adv_scheme
    use toolbox,     only : print_totalmass, escape_tm
    use global_data, only : mass_dat
    use zoom_tools,  only : update_parent
    use chem_param,  only : fscale, ntracet, ntrace, ntrace_chem
#ifdef MPI
    use mpi_const,   only : ntracetloc, myid, lmar, ntracet_ar
#endif
    use ParTools   , only : lmloc

    implicit none

    ! local
    real,dimension(:,:,:,:),pointer   :: rm, rxm, rym, rzm
    real,dimension(:,:,:),  pointer   :: m
    real,dimension(:,:,:,:),pointer   :: rmc
    integer                           :: region, n, offsetl, l, offsetn

    ! start

#ifdef MPI
    if ( ntracetloc == 0 ) return !WP! only PE's with nonzero ntracetloc
#endif

    do region=1,nregions

       m => mass_dat(region)%m_t
       rm => mass_dat(region)%rm_t
       rxm => mass_dat(region)%rxm_t
       rym => mass_dat(region)%rym_t
       rzm => mass_dat(region)%rzm_t

#ifdef MPI
       offsetn = 0
       if ( myid > 0 ) offsetn = sum(ntracet_ar(0:myid-1))
       do n=1,ntracetloc
          rm(1:im(region),1:jm(region),1:lm(region),n) = &
               (n+offsetn)*1e-9*m(1:im(region),1:jm(region),1:lm(region))/ &
               fscale(n+offsetn)
#else
       do n=1,ntracet
          rm(1:im(region),1:jm(region),1:lm(region),n) = &
               n*1e-9*m(1:im(region),1:jm(region),1:lm(region))/fscale(n)
#endif

          !   rm(1:10,1:10,1:10,n) = 40e-9*m(1:10,1:10,1:10)/fscale(n)
          if ( adv_scheme == 'slope' ) then
             rxm(:,:,:,n) = 0.0
             rym(:,:,:,n) = 0.0
             rzm(:,:,:,n) = 0.0
          end if
       end do

       ! for safety ...
       if ( .not. associated(chem_dat(region)%rm_k) ) then
         !write (*,'("ERROR - array chem_dat(region)%rm_k not allocated ...")')
         !write (*,'("ERROR in ",a)') 'sourcec_sinks/trace1' 
         !call escape_tm('Error in trace1')
         allocate( chem_dat(region)%rm_k(im(region),jm(region),lmloc, &
                                    ntracet+1:ntracet+ntrace_chem) )
       end if

       rmc => chem_dat(region)%rm_k
       offsetl = 0

#ifdef MPI
       if ( myid > 0 ) offsetl = sum(lmar(0:myid-1))
#endif

       do n=ntracet+1,ntrace 
#ifdef MPI
          do l=1,lmloc
#else
          do l=1,lm(region)
#endif
             rmc(1:im(region),1:jm(region),l,n) = &
                  n*1e-9*m(1:im(region),1:jm(region),l+offsetl)/fscale(n)
          end do
       end do

       nullify(m)
       nullify(rm)
       nullify(rxm)
       nullify(rym)
       nullify(rzm)
       nullify(rmc)

    end do

#ifdef MPI
    write (gol,*) 'trace1: rm initialized at mixing ratio :', (myid+1)*1e-9; call goPr
#else
    write (gol,*) 'trace1: rm initialized at mixing ratio :', 1e-9; call goPr
#endif

    do region = nregions,2,-1
       call update_parent(region)
    end do
    call print_totalmass(1)

  end subroutine trace1


  !--------------------------------------------------------------------- 
  ! 
  ! subroutine that is called after reading new fields 
  ! (clouds, surface winds, etc.)
  ! In this routine, 'chemistry' fields that depend on these data are calculated
  ! Routine is called from tracer.f90 (after reading data)
  ! 
  !--------------------------------------------------------------------- 

  subroutine trace_after_read
  
    ! --- begin ---------------------------------

    ! nothing to be done

  end subroutine trace_after_read


  !--------------------------------------------------------------------- 
  ! 
  !       source1
  !
  !
  !       purpose
  !       -------
  !       this subroutine changes the tracer mass and its slopes 
  !       by chemical sources
  !      
  !       interface
  !       ---------
  !       call source1
  !
  !       method
  !       ------
  !  
  !       
  !
  !       reference
  !       --------
  !       see above      
  !
  !---------------------------------------------------------------------

  subroutine source1(region)

    use dims,        only : itaur, tref, nsrce, idate, sec_month, sec_year, mlen
    use toolbox,     only : dumpfieldi
    use global_data, only : mass_dat, region_dat
    use datetime,    only : tau2date, get_day
#ifdef MPI
    use mpi_const,   only : myid
#endif

    implicit none

    ! input
    integer, intent(in)        :: region

    ! local
    real,dimension(:,:,:,:),pointer   :: rm, rxm, rym, rzm
    real,dimension(:,:,:),  pointer   :: m
    ! note gph is from 1-->lm+1 (since dec 2002) MK
    real,dimension(:,:,:),  pointer   :: gph
    real,dimension(:),      pointer   :: dxyp
    integer,dimension(:,:), pointer   :: zoomed
    integer,dimension(:,:), pointer   :: edge
    real                              :: dtime
    integer,dimension(6)              :: idater
    integer                           :: i, j, l, n
    ! conversion emission per month/year to per timestep
    real                              :: month2dt, year2dt
    integer                           :: ntim, day, sec_in_day, iseas, itim, mo

    ! start

    call tau2date(itaur(region),idater)
    ! emissions are added in two steps...XYZECCEZYX.
    dtime=float(nsrce)/(2*tref(region)) 
    ntim=86400/nint(dtime)    !number of timesteps in 24 hours.

    ! elapsed seconds this day
    sec_in_day = idater(4)*3600 + idater(5)*60 + idater(6)
    itim=sec_in_day/nint(dtime)+1                       !# time interval

    iseas=mod(idater(2)/3,4)+1 ! determine season
    mo=idate(2)                ! month

    month2dt=dtime/sec_month ! conversion to emission per timestep
    year2dt=dtime/sec_year

    day = get_day(idater(2),idater(3),mlen)

  end subroutine source1



  subroutine source2(region)
    ! ---------------------------------------------------------------------
    !
    !       source2
    !
    !
    !       purpose
    !       -------
    !       apply boundary conditions to selected tracers
    !      
    !       interface
    !       ---------
    !       call source2
    !
    !       method
    !       ------
    !  
    !       
    !       external
    !       --------
    !       none
    !
    !       reference
    !       --------
    !       see above      
    !                  
    !
    !---------------------------------------------------------------------

    use dims,        only : dx, dy, xref, yref, tref, nsrce
    use toolbox,     only : dumpfieldi,dumpfield
    use global_data, only : mass_dat,region_dat

    implicit none

    ! input
    integer, intent(in)           :: region

    ! local
    real,dimension(:,:,:,:),pointer   :: rm
    real,dimension(:,:,:),  pointer   :: m
    integer,dimension(:,:), pointer   :: zoomed
    integer,dimension(:,:), pointer   :: edge
    integer                           :: i, j, l, n
    real                              :: dtime, dxx, dyy

    ! start

    dtime=nsrce/(2*tref(region))  
    dxx=dx/xref(region)  !gridsize at current region
    dyy=dy/yref(region)  !gridsize at current region

  end subroutine source2


  !
  ! deallocate the memory and do everything needed to finalise chemistry
  !

  subroutine trace_end( msg, file_name, status )
  
    use dims,          only : region_name, im, jm, lm
    use dims         , only : idate
    use chem_param,    only : ntracet, ntrace_chem
    use global_data,   only : mass_dat
    use file_hdf     , only : THdfFile, Init, Done
    use io_hdf,        only : io_write
#ifdef with_budgets
    use budget_fine,   only : diagbud
    use budget_global, only : diagbudg
    use budget_global, only : init_mass
    use budget_global, only : sum_chemistry, sum_update, apply_budget_global
    use budget_global, only : sum_advection, sum_deposition, sum_stratosphere
#endif
    use emission,      only : sum_emission
#ifdef MPI
    use mpi_const, only : my_real, mpi_sum, com_lev
    use mpi_const, only : ntracetloc, lmloc, tracer_active, ierr
#endif
    use partools , only : myid, root_t, root_k
    use partools , only : Par_Gather_Tracer_k, Par_Barrier

    ! --- in/out ------------------------------------
    
    character(len=*), intent(in)  ::  file_name
    character(len=*), intent(in)  ::  msg
    integer, intent(out)          ::  status
    
    ! --- local -------------------------------------

    character(len=40)        :: frmt
    integer                  :: region, n, imr, jmr, lmr
    real                     :: mass, sum_chemistry_all, sum_wet_all
    real,dimension(:,:,:,:), pointer      :: rmc
    real,dimension(:,:,:,:), allocatable  :: fieldglob4d
    
    character(len=256)  ::  fname
    type(THdfFile)      ::  hdf

    ! --- begin --------------------------------------

    ! first output of the chemical fields:
    do region = 1,nregions

       imr = im(region) 
       jmr = jm(region) 
       lmr = lm(region)
       
       ! allocate global array on all PEs
       allocate(fieldglob4d(imr,jmr,lmr,ntrace_chem))
       
       ! tracers parallel over levels:
       rmc => chem_dat(region)%rm_k

       call Par_Gather_Tracer_k(fieldglob4d,imr,jmr,lmr,0,0,0,ntrace_chem,rmc,.false.,status)
       if (status/=0) then
         write (gol,'("error from par_gather_tracer_k; continue")'); call goPr
         exit  ! leave region loop
       end if

       if ( myid == root_k ) then

          ! create filename :  path/save_2001010100_glb3x2.hdf
          write (fname,'(a,"_",i4.4,3i2.2,"_",a,".hdf")') &
                   trim(file_name), idate(1:4), trim(region_name(region))
                   
          ! re-open save file:
          call Init( hdf, fname, 'write', status )
          if (status/=0) then
            write (gol,'("WARNING - error from start chemical output file:")'); call goPr
            write (gol,'("WARNING -   file  : ",a)') trim(fname); call goPr
            write (gol,'("WARNING -   skip and continue ...")'); call goPr
            cycle  ! next region
          end if

          ! write tracers:
          call io_write( hdf%id, imr,'LON'//region_name(region),&
               jmr,'LAT'//region_name(region), &
               lmr,'HYBRID',ntrace_chem,'NTRACE_CHEM',fieldglob4d,'rmc')
          
          ! close save file:
          call Done( hdf, status )
          if (status/=0) then
            write (gol,'("WARNING - error from end chemical output file:")'); call goPr
            write (gol,'("WARNING -   file  : ",a)') trim(file_name)//region_name(region); call goPr
            write (gol,'("WARNING -   skip and continue ...")'); call goPr
            cycle  ! next region
          end if

       end if
       call Par_Barrier

       nullify(rmc)      
       deallocate(fieldglob4d)

    end do

#ifdef with_budgets
    if (apply_budget_global) then

      if ( myid == root_t ) then 
        write (gol,'("--------------------------------------------------------")'); call goPr
        write (gol,'("trace_end: Budget of tracer 1 in all regions (kg) :     ")'); call goPr
        write (gol,'("--------------------------------------------------------")'); call goPr
      end if

      do region =1,nregions 

        call diagbud(region,3)    !add depdry
        call diagbudg(region,3)    !add depdry
        call diagbud(region,2)    !write output
        call diagbudg(region,2)    !write output, free memory...

        mass = sum(mass_dat(region)%rm_t(1:im(region),1:jm(region),1:lm(region),1))

        call get_format(mass,frmt)

        if ( myid == root_t ) then 
          if (apply_budget_global) then
            write (gol,'("Region             ",i3)') region; call goPr
            write (gol,frmt) 'Initial mass   :            ',init_mass(region); call goPr
            write (gol,frmt) 'mass emitted   :            ',sum_emission(region); call goPr
            write (gol,frmt) 'chemistry      :            ',sum_chemistry(region); call goPr
            write (gol,frmt) 'update_p       :            ',sum_update(region); call goPr
            write (gol,frmt) 'advection      :            ',sum_advection(region); call goPr
            write (gol,frmt) 'deposition     :            ',sum_deposition(region); call goPr
            write (gol,frmt) 'stratosphere   :            ',sum_stratosphere(region); call goPr
            write (gol,frmt) 'Final mass     :            ',mass; call goPr
            write (gol,frmt) 'Budget (error) :            ', &
                 init_mass(region)  + sum_emission(region) &
                 +sum_advection(region) + sum_chemistry(region) &
                 +sum_update(region)  &
                 +sum_deposition(region) + sum_stratosphere(region) - mass; call goPr
            write (gol,'("--------------------------------------------------------")'); call goPr
          end if
        end if

      end do
      
    endif  !apply_budget_global
#endif

    ! ok
    status = 0

  end subroutine trace_end


  ! ***

  
  subroutine get_format(mm,frmt)

    real,intent(in)              :: mm
    character(len=*),intent(out) :: frmt

    integer :: is

    if ( mm > 0.0 ) then
      is = int(alog10(mm))
    else
      is = 0
    end if

    select case(is)
      case(1:10)
        write (frmt,'(a9,i1,a1)') '(a30,f14.', 10-is, ')'
      case default
        frmt =  '(a30,1pe12.5)'
    end select

  end subroutine get_format


end module sources_sinks
