!-----------------------------------------------------------------------
!
!**** Global Atmospheric Tracer Model TM5
!
! purpose
! -------
! Solves the tracer continuitiy equation on an eulerian grid for
! an arbitrary no. of tracers.
! Allow zoom regions with higher resolution.
! Perform advection, vertical transport, emissions, chemistry, deposition, ....
!
! interface
! ---------
! main program
!
! method
! ------
! Initialize the model and then cycle continuously through
! the main model time loop.
!
! externals
! ---------
!
! reference
! --------- 
! See the main website :  www.phys.uu.nl/~tm5/
!
!-----------------------------------------------------------------------

program Tracer

  use GO              , only : TDate, NewDate, IncrDate, wrtgol
  use GO              , only : rTotal, operator(-), operator(+), operator(==), operator(>)
  use GO              , only : GO_Print_Init, GO_Print_Done, gol, goPr
  use GO              , only : TrcFile, Init, Done, ReadRc
  use Dims            , only : nregions
  use Dims            , only : region_status => status
  use Dims            , only : revert, staggered
  use Dims            , only : newsrun, newmonth
  use Dims            , only : itau, itaui, itaue, itaur
  use Dims            , only : idate, idatee
  use Dims            , only : nread, ndyn, nconv, nsrce, nchem
  use Dims            , only : ndyn_max
  use datetime        , only : inctime, tstamp
  use ParTools        , only : Par_Init, Par_Done
  use ParTools        , only : Par_Barrier
  use ParTools        , only : myid, root, npes
  use tracer_data     , only : Par_Check_Domain
  use global_data     , only : rcfile
  use Meteo           , only : Meteo_Init, Meteo_Done
  use Meteo           , only : Meteo_Setup_Mass, Meteo_Setup_Other
  use Sources_Sinks   , only : trace0, trace_after_read
  use Zoom_Tools      , only : coarsen_region
  use Advect          , only : dynam0
  use AdvectM_CFL     , only : Check_CFL, Setup_MassFlow
  use Diffusion       , only : Calc_Kzz
  use ModelIntegration, only : Proces_Init, Proces_Done, Proces_Region
  use ModelIntegration, only : advect_apply, convec_diffusion_apply
  use InitExit        , only : start, exitus
  use Toolbox         , only : escape_tm
  use user_output,      only : user_output_mean

  implicit none
 
  ! --- const ---------------------------------------
  
  character(len=*), parameter  ::  pname = 'tracer'
   
  ! --- local ---------------------------------------

  integer           ::  region
  integer           ::  status
  integer           ::  n
  integer           ::  idate1(6), idate2(6)
  type(TDate)       ::  tread1, tread2
  type(TDate)       ::  tdyn, tend, tr(2)
  logical           ::  isfirst
  integer           ::  nhalf

  type(TrcFile)     ::  rcF
  logical           ::  go_print_all
  logical           ::  go_print_apply
  logical           ::  go_print_trace
  logical           ::  go_print_prompt_pe
  
  ! --- begin ---------------------------------------
  
  !---------------!
  ! model startup !
  !---------------!
  
  ! ~~~ par
  
  ! init parallelisation
  call Par_Init( status )
  if (status/=0) call escape_tm('ERROR in tracer')

  ! ~~~ messages
  
  ! read settings:
  call Init( rcF, rcfile, status )
  if (status/=0) call escape_tm('ERROR in tracer')
  call ReadRc( rcF, 'go.print.all', go_print_all, status, default=.false. )
  if (status/=0) call escape_tm('ERROR in tracer')
  call ReadRc( rcF, 'go.print.prompt.pe', go_print_prompt_pe, status, default=npes>1 )
  if (status/=0) call escape_tm('ERROR in tracer')
  call ReadRc( rcF, 'go.print.trace', go_print_trace, status, default=.false. )
  if (status/=0) call escape_tm('ERROR in tracer')
  call Done( rcF, status )
  if (status/=0) call escape_tm('ERROR in tracer')
    
  ! standard output by root only:
  go_print_apply = go_print_all .or. (myid==root)
  
  ! setup standard output processing:
  call GO_Print_Init( status, &
                        apply=go_print_apply, trace=go_print_trace, &
                        prompt_pe=go_print_prompt_pe, pe=myid )
  if (status/=0) call escape_tm('ERROR in tracer')
  
  write (gol,'("")'); call goPr
  write (gol,'("*************************************************************")'); call goPr
  write (gol,'("***                                                       ***")'); call goPr
  write (gol,'("***                                                       ***")'); call goPr
  write (gol,'("***        Global Atmospheric Tracer Model TM5            ***")'); call goPr
  write (gol,'("***                                                       ***")'); call goPr
  write (gol,'("***                                                       ***")'); call goPr
  write (gol,'("*************************************************************")'); call goPr
  write (gol,'("")'); call goPr

  ! ~~~ meteo  

  ! setup meteo input:
  write (gol,'("setup meteo ...")'); call goPr
  call Meteo_Init( status )
  if (status/=0) call escape_tm('ERROR in tracer')

  ! init processes
  write (gol,'("setup processes ...")'); call goPr
  call Proces_Init( status )
  if (status/=0) call escape_tm('ERROR in tracer')

  ! ~~~ other
  
  ! set-up and read user input; 
  ! return time interval for which meteo was read:
  write (gol,'("start ...")'); call goPr
  call Start( tread1, tread2, status )
  if (status/=0) call escape_tm('ERROR in tracer')  

  !-----------!
  ! main loop !
  !-----------!
  
  ! current time (begin of dynamics step)
  tdyn = NewDate( time6=idate  )
  tend = NewDate( time6=idatee )

  ! first step in time loop ?
  isfirst = .true.
  nhalf   = 0

  ! time loop over steps ndyn/2 (!)
  do 

    ! end time reached ? then leave
    if ( (revert*itau) >= (revert*itaue) ) exit
    
    ! next half step
    nhalf = modulo(nhalf,2) + 1

    !
    ! *** info ***
    !

    ! display current time interval:
    if ( nhalf == 1 ) then
      write (gol,'(" ")'); call goPr
      call wrtgol( '>>> dynamics step from : ', tdyn ); call goPr
      write (gol,'(" ")'); call goPr
    end if

    !
    ! *** setup data ***
    !

    ! first set-up of chemistry related work...repeated if newmonth
    if ( newmonth .and. (.not. newsrun) ) then
       call trace0( status )
       if (status/=0) call escape_tm('ERROR in tracer')
    end if

    ! reached end of time interval for which meteo is valid ? then setup new meteo:
    if ( tdyn == tread2 ) then

      ! reset possible reduced timestep due to CFL:
      ndyn =  ndyn_max
      nsrce = ndyn_max
      nconv = ndyn_max
      nchem = ndyn_max

      !in zoom DO this for all zoom regions,now for only one!
      do region = 1,nregions
         call Par_Check_Domain( region, 'n', 'tracer' )
      end do

      ! setup meteo data for next interval;
      ! nread is the length (in seconds) of the interval in which 
      ! surface pressure is interpolated (and mass fluxes are constant)
      tread1 = tdyn
      tread2 = tdyn + IncrDate(sec=nread)
      if ( tread2 > tend ) tread2 = tend

      ! n is the number of dynamic intervals within the
      ! time interval for which the meteo has been setup:
      n = ceiling( rTotal(tread2-tread1,'sec') / real(ndyn) )
      ndyn = nint( rTotal(tread2-tread1,'sec') / n )

      ! setup meteo:
      !  o skip first time; already called in 'initexit/start'
      !  o check pressure implied by advection if advection is applied
      call Meteo_Setup_Mass( tread1, tread2, status, check_pressure=advect_apply )
      if (status/=0) call escape_tm('ERROR in tracer') 

      ! determine dynamic timestep ndyn for this interval [tread1,tread2] ;
      ! the initial number of time steps n is increased until no cfl
      ! violations occure
      call Check_CFL( tread1, tread2, n, status )
      if (status/=0) call escape_tm('ERROR in tracer') 

    end if
    
    ! setup meteo for dynamic step tdyn+[0,ndyn]
    if ( nhalf == 1 ) then
      ! time range of dynamic step:
      tr(1) = tdyn
      tr(2) = tdyn + IncrDate( sec=ndyn )
      ! convert pu/pv to am/bm/cm, eventually time interpolated
      call Setup_MassFlow( tr, status )
      if (status/=0) call escape_tm('ERROR in tracer') 
      ! setup (interpolate?) other meteo:
      call Meteo_Setup_Other( tr(1), tr(2), status )
      if (status/=0) call escape_tm('ERROR in tracer') 

      ! recalculate diffusion coeffiecents if necessary
      if ( convec_diffusion_apply ) call Calc_Kzz

      ! recalculate 'chemistry' fields if necessary
      call trace_after_read

    end if

    !
    ! *** processes ***
    !
    
    if ( ndyn > 0 ) then

      ! reset the process status counters:
      if ( nhalf == 1 ) region_status(1:nregions) = 0

      ! info ...
      !if (myid==root) write (*,'(a,": Start processing main region")') pname
 
      tr(1) = tdyn
      tr(2) = tdyn + IncrDate(sec=ndyn/2)
      if ( nhalf == 1 ) then
        write (gol,'(" ")'); call goPr
        call wrtgol( '--> first  half : ', tr(1), ' - ', tr(2) ); call goPr
        write (gol,'(" ")'); call goPr
      end if
      if ( nhalf == 2 ) then
        write (gol,'(" ")'); call goPr
        call wrtgol( '--> second half : ', tr(1), ' - ', tr(2) ); call goPr
        write (gol,'(" ")'); call goPr
      end if

      !synchronize time-count regions....
      itaur(:) = itau 
      call Par_Barrier

      !
      ! start recursive process for the main region = 1
      ! Within the advection, all the other processes are incorporated...
      ! Also the regions 'below' are called recursively....
      !

      ! start processing main region
      call Proces_Region( 1, tr, status )
      if (status/=0) call escape_tm('ERROR in tracer') 

      ! check times ...
      do region = 2, nregions
        if ( itaur(region) /= itaur(1) ) then
          call escape_tm( 'tracer: exit of routine proces_region with non-synchronized clocks' )
        end if
      end do

    end if
    
    !
    ! *** next ***
    !

    ! advance the model time with ndyn/2 seconds:
    call inctime
    tdyn = tdyn + IncrDate( sec=nint(ndyn/2.0) )

    if (mod(itau,ndyn_max) == 0) then
        call user_output_mean
    endif

    ! not the first step in time loop any more ...
    isfirst = .false.

  end do   ! main loop 

  !---------------!
  ! model finish  !
  !---------------!
  
  ! done processes
  call Proces_Done( status )
  if (status/=0) call escape_tm('ERROR in tracer')

  ! close meteo files etc
  call Meteo_Done( status )
  if (status/=0) call escape_tm('ERROR in tracer')

  ! done parallelisation
  call Par_Done( status )
  if (status/=0) call escape_tm('ERROR in tracer')

  ! done standard output:
  call GO_Print_Done( status )
  if (status/=0) call escape_tm('ERROR in tracer')
  
  ! store save file etc
  call exitus

end program Tracer

