#! /bin/csh -ef
# --- init -----------------------------

# set program name and location: 
if ( $0 =~ /* ) then
  set script = $0
else
  set script = $PWD/$0
endif
set scriptdir = ${script:h}
set prog = ${script:t}

source ${scriptdir}/platform_specific_settings.csh

# --- help -----------------------------

if ( ${#argv} < 1 ) then
cat << EOF

NAME
  $prog - run TM5 model
  
USAGE
  $prog <rcfile>
  
RCFILE

  See the template 'TM5/rcfiles/tm5.rc' .
  
EOF
exit(0)
endif

# --- arguments ------------------------

set rcfile      = $1
set time_start      = $2
set time_final      = $3
set compdir = ${scriptdir}/..

# --- binaries ----------
set basename = /usr/bin/basename
if ( ! -x $basename ) then
    set basename = /bin/basename
endif
if ( ! -x $basename ) then
    echo "${prog} - ERROR!  Cannot find basename executable in /usr/bin or /bin"
    exit 1
endif

# --- begin -----------------------------------------------------------
set iserror = 'F'


echo "$prog - Read settings from ${rcfile} ..."
set rundir   = `$scriptdir/go_readrc $rcfile 'rundir'`
set outputfreq   = `$scriptdir/go_readrc $rcfile 'outputfreq'`
set nlag   = `$scriptdir/go_readrc $rcfile 'nlag'`
set regtype   = `$scriptdir/go_readrc $rcfile 'regtype'`

set with_mpi = `$scriptdir/go_readrc $rcfile 'with.mpi'`
set npes     = `$scriptdir/go_readrc $rcfile 'npes'`
if ( "${with_mpi}" != "T" ) set npes = 1
set inputdir = `$scriptdir/go_readrc $rcfile 'inputdir'`
set istart = `$scriptdir/go_readrc $rundir/tm5_runtime.rc 'istart'`
set sitesrc   = `$scriptdir/go_readrc $rcfile 'sites.rc'`
set obsdir   = `$scriptdir/go_readrc $rcfile 'obs.dir' 'NONE'`
if( "$obsdir" == "NONE") then
    set obsdir = `${dirname} $sitesrc`
endif


# ----------------
# check settings of rc file versus those in source code used
# ----------------

cd $compdir/src
if($with_mpi == 'T') then
   # check the NPES setting in mpi_const:
   set npesf = `grep -i 'npes =' mpi_const.F90 | awk '{print $5}'`
   echo $npes $npesf
   if ($npes != $npesf) then
    echo "${prog} - ERROR!  npes ($npes) in rc file ($rcfile) should be equal to npes ($npesf) in mpi_const.F90"
    set iserror = 'T'
   endif
   set ntracet = `grep -i 'ntracet =' chem_param.F90 | awk '{print $6}'`
   @ tt = $ntracet % $npes  
   if($tt != 0 ) then
    echo "${prog} - ERROR!  ntracet ($ntracet) in rc file ($rcfile) should be integer multiple of npes ($npes)"
    set iserror = 'T'
   endif
endif

# check that start.state is a bare filename--no directory.
set start_state = `$scriptdir/go_readrc $rcfile 'start.state'`
if ( `${basename} $start_state` != $start_state) then
    echo "${prog} - ERROR!  start.state in $rcfile apparently has a directory specification."
    echo "${prog}   This is deprecated in favor of a bare filename, e.g. 'start.state: savestate.hdf'"
    echo "${prog}   Value in file: $start_state"
    set iserror = 'T'
endif

# check the nlag setting in enkf_constants:
set nlagf = `sed 's/\\!.*//g' enkf_constants.F90 | grep -i 'nlag[      ]*=' | awk -F\: '{print $3}' | awk -F= '{print $2}' | tr -d "'" `
# anatomy:  filter all comments                   get assignment only        get part after ::       get part after =       strip single quotes
if ($nlagf != $nlag) then
    echo "${prog} - ERROR!  nlag ($nlag) in rc file ($rcfile) should be equal to nlag ($nlagf) in enkf_constants.F90"
    set iserror = 'T'
endif

# check the timestep setting in enkf_constants:
set outputfreqf = `sed 's/\\!.*//g' enkf_constants.F90 | grep -i 'timestep[      ]*=' | awk -F\: '{print $3}' | awk -F= '{print $2}' | tr -d "'" `
# anatomy:        filter all comments                   get assignment only            get part after ::       get part after =       strip single quotes
if ($outputfreqf != $outputfreq) then
    echo "${prog} - ERROR!  outputfreq ($outputfreq) in rc file ($rcfile) should be equal to timestep ($outputfreqf) in enkf_constants.F90"
    set iserror = 'T'
endif

# check the regtype setting in enkf_constants:
set regtypef = `sed 's/\\!.*//g' enkf_constants.F90 | grep -i 'regtype[      ]*=' | awk -F\: '{print $3}' | awk -F= '{print $2}' | tr -d "'" `
# anatomy:     filter all comments                   get assignment only           get part after ::       get part after =       strip single quotes
if ($regtypef != $regtype) then
    echo "${prog} - ERROR!  regtype ($regtype) in rc file ($rcfile) should be equal to regtype ($regtypef) in enkf_constants.F90"
    set iserror = 'T'
endif

if ( $iserror == 'T' ) then
    echo "${prog} - Please fix error(s) above.  Exiting prematurely."
    exit 1
endif

# link input
#

cd $rundir

echo "$prog - link input files ..."

set transcomdir   = `$scriptdir/go_readrc $rcfile 'transcomdir'`
set dir1x1   = `$scriptdir/go_readrc $rcfile 'dir1x1'`
#set deltapco2_prefix   = `$scriptdir/go_readrc $rcfile 'deltapco2.prefix'`
#set bio_covariance =  `$scriptdir/go_readrc $rcfile 'bio.covariance'`
#set ocn_covariance =  `$scriptdir/go_readrc $rcfile 'ocn.covariance'`
set regtype =  `$scriptdir/go_readrc $rcfile 'regtype'`
    
# general input
#/bin/cp $inputdir/station.list .
/bin/cp $inputdir/RedGrid .
/bin/cp $inputdir/fintwb.data  .
# 1x1 files needed
#/bin/ln -sf $dir1x1/co2_init_new.hdf co2_init.hdf
/bin/ln -sf $dir1x1/co2_init_opt.hdf co2_init.hdf
/bin/ln -sf $dir1x1/landfraction.hdf landfraction.hdf
/bin/ln -sf $dir1x1/distances.hdf .
/bin/ln -sf $dir1x1/transcom_${regtype}.hdf .
#/bin/ln -sf $dir1x1/${bio_covariance} .
#/bin/ln -sf $dir1x1/${ocn_covariance} .
# flux input from TC Continuous
#/bin/ln -sf $transcomdir/transcom_sib_hflux*.hdf .
#/bin/ln -sf $transcomdir/transcom_sib_dflux*.hdf .
#/bin/ln -sf $transcomdir/transcom_casa_dflux*.hdf .
#/bin/ln -sf $dir1x1/transcom_ff95.hdf .
#/bin/ln -sf $transcomdir/transcom_taka02_flux*.hdf .
#/bin/ln -sf $dir1x1/${deltapco2_prefix}* .
#/bin/ln -sf $dir1x1/bioparams* .
#/bin/ln -sf $dir1x1/biofireparams* .
#/bin/ln -sf $dir1x1/BBguido* .
# fossil fuel inputs
#/bin/ln -sf $dir1x1/ff_input/co2ff.edgar_cdiac.seasnam.1x1.hdf* .
#flux input from old TC
#ln -sf $dir1x1/transcom_ff95.hdf transcom_ff95.hdf
#ln -sf $dir1x1/transcom_ocean.hdf transcom_ocean.hdf
#ln -sf $dir1x1/transcom_nep.hdf transcom_nep.hdf

# site selection

#set blacklist = (xxx)

set idate = `echo ${time_start} | /usr/bin/cut -c 1-8`
set idate_final = `${scriptdir}/date_add ${idate} $outputfreq`

if (-f obs_final.list) rm -f obs_final.list
if (-f obs_forecast.list) rm -f obs_forecast.list
if (-f obs_forecast_preBL.list) rm -f obs_forecast_preBL.list

echo $obsdir

# range of times to extract for final obs
set idate = `echo ${time_start} | /usr/bin/cut -c 1-8`
@ increment = $outputfreq - 1 # note that this is now one day less than outputfreq because we're going to add 23:59:00 to it.
set idate_final = `${scriptdir}/date_add ${idate} $increment`

# time_start has YYYYMMDDHH, no minute or second
set t0 = ( `echo ${time_start} | awk '{print substr($0,1,4), substr($0,5,2), substr($0,7,2), substr($0,9,2);}'` 00 00 UTC )
set t1 = ( `echo ${idate_final} | awk '{print substr($0,1,4), substr($0,5,2), substr($0,7,2);}'` 23 59 00 )
set dt = $outputfreq # in days

# read the list of site categories to be used for the obs_final.list file
set n_obs_final = `$scriptdir/go_readrc $sitesrc 'n.obs.final'`
@ icomp = 0
set components = ()
echo "$prog - obs_final.list is constructed from $n_obs_final components:"
while ( $icomp < $n_obs_final) 
    @ icomp = $icomp + 1
    set component = `$scriptdir/go_readrc $sitesrc "obs.final.$icomp"`
    set components = ($components $component)
    echo "  [$icomp] " $component
   if (! -d `${dirname} ${sitesrc}`/`${dirname} ${component}` ) then
        echo "${prog} - Cannot find final obs directory for `${dirname} ${sitesrc}`/`${dirname} ${component}`"
        echo "${prog} - Exiting."
        exit 1
    endif
end

    foreach component ( $components )

	set obscompfile = ${obsdir}/${component}

	if (-e ${obscompfile}) then
        # awk code explanation: 
        #
        # The first if-clause extracts only the date range we want;
        # the second if-clause filters for thresholds of latitude,
        # longitude, and height.  The second clause uses a special
        # syntax to either output a valid line to the obs list file,
        # or emit an error message if the line has invalid lat, lon,
        # or height.
        #
        # The line-continuation backslashes need some explanation; see:
        # http://www.gnu.org/software/gawk/manual/gawk.html#Statements_002fLines
        # Specifically, the comment that:
        #
        # Caution: Backslash continuation does not work as described
        # with the C shell. It works for awk programs in files and for
        # one-shot programs, provided you are using a POSIX-compliant
        # shell, such as the Unix Bourne shell or bash. But the C
        # shell behaves differently! There, you must use two
        # backslashes in a row, followed by a newline. Note also that
        # when using the C shell, every newline in your awk program
        # must be escaped with a backslash.
        #
        # In this case, the 'program' is that part contained in single
        # quotes.  The set of -v arguments gets a regular
        # line-continuation backslash.

        awk -v t0="$t0"  -v dt=${dt} \
            'BEGIN { pt0=mktime(t0);pt1=pt0+86400*dt;} \\
            { pt=mktime($5 " " $6 " " $7 " " $8 " " $9 " " $10 " UTC"); \\
            if (pt >= pt0 && pt < pt1) { \\
                if (($2 >= -90) && ($2 <= 90) && \\
                 ($3 >= -180) && ($3 <= 180) && \\
                 ($4 >= 0) && ($4 <= 200000)) \\
                {print $0 >> "obs_final.list"} else {print "WARNING! Filtered invalid FINAL observation: " $0}}}' ${obscompfile}

    else
        echo "WARNING:  ${obscompfile} does not exist"
    endif

end # foreach
set t0f = `echo $t0|awk '{print $1"-"$2"-"$3" "$4":"$5":"$6" UTC";}'`
set t1f = `echo $t1|awk '{print $1"-"$2"-"$3" "$4":"$5":"$6" UTC";}'`
set nfinal = 0
if (-e obs_final.list) then
    set nfinal = `cat obs_final.list|wc -l`
endif
echo "${prog} - obs_final.list contains $nfinal obs between $t0f and $t1f"

# range of times to extract depends on whether we just started the
# filter or not (i.e., are in the very first assimilation cycle, in
# which we want obs from 2000-01-01 to 2000-02-05 when using a 5-week
# lag)

set nstep = ${nlag}
@ dt = $outputfreq # in days
if ( $istart == 6 ) then
    set nstep = 1
    @ dt = $outputfreq * $nlag
endif
@ int2 =  ($nstep - 1) * $outputfreq
set idates = `echo ${time_start} | /usr/bin/cut -c 1-8`
set idate = `${scriptdir}/date_add ${idates} $int2`
@ int2 =  $nlag * $outputfreq
set idate_final = `${scriptdir}/date_add ${idates} $int2`

set t0 = ( `echo ${idate} | awk '{print substr($0,1,4), substr($0,5,2), substr($0,7,2);}'` 00 00 00 )
set t1 = ( `echo ${idate_final} | awk '{print substr($0,1,4), substr($0,5,2), substr($0,7,2);}'` 23 59 00 )


# read the list of site categories to be used for the obs_forecast.list file
set n_obs_forecast = `$scriptdir/go_readrc $sitesrc 'n.obs.forecast'`
@ icomp = 0
set components = ()
echo "$prog - obs_forecast.list is constructed from $n_obs_forecast components:"
while ( $icomp < $n_obs_forecast) 
    @ icomp = $icomp + 1
    set component = `$scriptdir/go_readrc $sitesrc "obs.forecast.$icomp"`
    set components = ($components $component)
    echo "  [$icomp] " $component
    if (! -d `${dirname} ${sitesrc}`/`${dirname} ${component}` ) then
        echo "${prog} - Cannot find forecast obs directory for `${dirname} ${sitesrc}`/`${dirname} ${component}`"
        echo "${prog} - Exiting."
        exit 1
    endif
end

# now make list with obs forecasted (that is, *used* for filter optimization)
foreach component ( $components )

    set obscompfile = ${obsdir}/${component}

    if (-e ${obscompfile}) then

        # awk code explanation: 
        #
        # The first if-clause extracts only the date range we want;
        # the second if-clause filters for thresholds of latitude,
        # longitude, and height.  The second clause uses a special
        # syntax to either output a valid line to the obs list file,
        # or emit an error message if the line has invalid lat, lon,
        # or height.
        #
        # The line-continuation backslashes need some explanation; see:
        # http://www.gnu.org/software/gawk/manual/gawk.html#Statements_002fLines
        # Specifically, the comment that:
        #
        # Caution: Backslash continuation does not work as described
        # with the C shell. It works for awk programs in files and for
        # one-shot programs, provided you are using a POSIX-compliant
        # shell, such as the Unix Bourne shell or bash. But the C
        # shell behaves differently! There, you must use two
        # backslashes in a row, followed by a newline. Note also that
        # when using the C shell, every newline in your awk program
        # must be escaped with a backslash.
        #
        # In this case, the 'program' is that part contained in single
        # quotes.  The set of -v arguments gets a regular
        # line-continuation backslash.

        awk -v t0="$t0"  -v dt=${dt} \
            'BEGIN { pt0=mktime(t0);pt1=pt0+86400*dt;} \\
            { pt=mktime($5 " " $6 " " $7 " " $8 " " $9 " " $10 " UTC"); \\
            if (pt >= pt0 && pt < pt1) { \\
                if (($2 >= -90) && ($2 <= 90) && \\
                 ($3 >= -180) && ($3 <= 180) && \\
                 ($4 >= 0) && ($4 <= 200000)) \\
                {print $0 >> "obs_forecast.list"} else {print "WARNING!  Filtered invalid FORECAST observation: " $0}}}' ${obscompfile}
    else


# now mao "WARNING:  ${obscompfile} does not exist"
    endif

end # foreach

#rm -f blacklist.txt
#cp $scriptdir/select_obs.awk .
#foreach stat ($blacklist)
#   awk -f select_obs.awk f1=$stat obs_forecast_preBL.list > obs.list
#   cp obs.list obs_forecast_preBL.list
#end
#mv obs_forecast_preBL.list obs_forecast.list
set t0f = `echo $t0|awk '{print $1"-"$2"-"$3" "$4":"$5":"$6" UTC";}'`
set t1f = `echo $t1|awk '{print $1"-"$2"-"$3" "$4":"$5":"$6" UTC";}'`
set nfcast = 0
if (-e obs_final.list) then
    set nfcast = `cat obs_forecast.list|wc -l`
endif
echo "${prog} - obs_forecast.list contains $nfcast obs between $t0f and $t1f"

# leave the obs lists behind in archive-friendly format
set idate = `echo ${time_start} | /usr/bin/cut -c 1-8`
cp obs_forecast.list obs_forecast.list.${idate}
cp obs_final.list obs_final.list.${idate}
if (-e pseudo.list) then 
  cp pseudo.list pseudo.list.${idate}
endif

# In forward mode, we need hdf files with parameters.  These are
# assumed to be called "xac_01" and located in a file called
# "savestate.hdf".  There should be a different savestate.hdf for each
# week, located in a subdirectory named using the 8-digit idate.  The
# rcfile key "fwd.parms.dir" gives the location of the parent of these subdirs.
# For instance, if fwd.parms.dur = /p73/co2/andy/enso/output, then
# we expect hdf files of the form /p73/co2/andy/enso/output/20010101/savestate.hdf.
#
# Parameters are read by the routine "ReadParams" in enkf_common,
# which is called by "start" in initexit.F90.
#
set fwd_parms_dir = `$scriptdir/go_readrc $rcfile 'fwd.parms.dir' missing`
echo ${fwd_parms_dir}
if ( "$fwd_parms_dir" != "missing" ) then
    set source_file = ${fwd_parms_dir}/${idate}/savestate.hdf
    echo ${source_file}
    if (! -e ${source_file} ) then
	echo "${prog} - Cannot find ${source_file}"
	echo "${prog} - Is fwd.parms.dir value of ${fwd_parms_dir} correct in rc file?"
	echo "${prog} - Exiting."
	exit 1
    endif
    ln -sf ${source_file} fwd_parms.hdf
endif
# --- end --------------------------------------------------------------
