module chemloss_ch4_ohcl
  !-----------------------------------------------------------------------
  !       purpose
  !       -------
  !       perform ch4 photochemical loss from precalculated fields 
  !
  !       interface
  !       ---------
  !       call chemloss_apply
  !
  !       method
  !       ------
  !       subroutine is called from its general parent routine apply_chemloss
  !-----------------------------------------------------------------------
  use Dims          
  use global_types,  only: d3_data
  use emission_data, only: gridbox_per_month_to_flux
  use chem_param
  implicit none

  private

  ! public routines:
  public   :: kohcl
  public   :: declare_chemloss_ch4_ohcl
  public   :: free_chemloss_ch4_ohcl
  public   :: calc_chemloss_ch4_ohcl
  public   :: interp_loss

  type(d3_data), dimension(nregions), target  :: kohcl       !  CH4 loss [1/timestep]


contains

  subroutine declare_chemloss_ch4_ohcl

  implicit none
  integer                                      :: status, region, imr,jmr, lmr

     do region=1,nregions
      imr=im(region) ;jmr=jm(region) ;lmr=lm(region)
      allocate(kohcl(region)%d3(imr,jmr,lmr))  ! oh+ch4, cl+ch4 sink
      kohcl(region)%d3(:,:,:) = 0.0
     enddo


  end subroutine declare_chemloss_ch4_ohcl

  subroutine calc_chemloss_ch4_ohcl

#ifdef MPI
  use mpi_const
  use mpi_comm,    only: barrier
#endif
  use dims,  only : nlon60, nlat45, tref, nchem
  use io_save,     only: readtm3hdf_old
  use emission_data,	only: chardate
  use chem_param, only: xmc
  use global_data, only: inputdir_ch4

  implicit none
  integer  :: i_1,j_1,i_10,j_10    
  integer  :: i, j, l,ii, n, nmonth    
  integer  :: io, sfstart, ifail, sfend    
  real     :: year2s, month2s, dtime    
  real     :: lat,lon
  integer                                      :: status
  integer                                      :: region, imr, jmr, lmr
  integer, parameter                           :: annual_field = 0, rank2 = 2, level1 = 1
  character(len=10)                            :: cdate
  real                                         :: year2month
  character (len=55)			       :: filename
  character (len=5) 			       :: varname    
  character (len=8) 			       :: recname
  character(len=2),dimension(12),parameter     :: mlab=(/'01','02','03','04','05','06','07','08','09','10','11','12'/)   
  character(len=5),parameter :: l_lab='k_ch4'
  real,dimension(:,:,:),allocatable    :: field3d_in,field3d_out
  integer 	:: irec

   if (newmonth) then

     allocate(field3d_out(nlon60,nlat45,lm(1)))
     allocate(field3d_in(nlon60,nlat45,25))

#ifdef MPI
!     if (myid==root) then
#endif
     filename=trim(inputdir_ch4)//'/k_CH4.hdf'
     recname=trim(l_lab)//'_'//trim(mlab(idate(2)))
     print *,filename,recname
     call readtm3hdf_old(filename,3,nlon60,nlat45,25,idate(2)-1,field3d_in,recname) 
     region=1
!    call interpolation if necessary
     if (lm(1).gt.25 ) then 
      print *,'calling interpolation'
      call interp_loss(field3d_in,field3d_out)
     else
      field3d_out=field3d_in
     endif
     dtime=float(nchem)/(2*tref(region))
     kohcl(1)%d3(:,:,:)=(1.-exp(-field3d_out(:,:,:)*dtime))

#ifdef MPI
!     endif
#endif
    ! use global sink file also for zoom regions

    do region=2,nregions
   
      dtime=float(nchem)/(2*tref(region))
      i_10 = int((xbeg(region) - xbeg(1)) / dx )
      j_10 = int((ybeg(region) - ybeg(1)) / dy )

      do i=1, im(region)

        ! calculate corresponding i index (i_1) for global region 1
        i_1 = i_10 + int(i/xref(region))


        do j=1, jm(region)

          ! calculate corresponding j index (j_1) for global region 1

          j_1 = j_10 + int(j/yref(region))

          do l=1, lm(region)
            kohcl(region)%d3(i,j,l) = (1.0-exp(-field3d_out(i_1, j_1, l) * dtime))
          enddo

        enddo
      enddo

   enddo
   deallocate(field3d_in)
   deallocate(field3d_out)

   endif

  
  end subroutine calc_chemloss_ch4_ohcl

  subroutine interp_loss(field3d_in,field3d_out)
#ifdef MPI
  use mpi_const
  use mpi_comm,    only: barrier
#endif
  use dims,  only : nlon60, nlat45, nlon360, nlat180
  use global_data, only: mass_dat,region_dat
  use meteodata  , only : lli, levi, sp_dat
  use grid       , only : TllGridInfo, TLevelInfo, Init, Done, Fill3D

  implicit none
  real, parameter  :: a_in(0:25)  = (/&
       0.0000000, 7.3677430, 210.39389, 855.36176,  2063.7798,  3850.9133,  6144.3149,             &
       8802.3564,  11632.759, 14411.124, 16899.469, 18864.750, 20097.402, 20429.863, 19755.109,    &
       18045.184, 15379.806, 12077.446, 8765.0537, 6018.0195, 3960.2915, 1680.6403, 713.21808,     &
       298.49579, 95.636963, 0.0000000/)

  real, parameter  :: b_in(0:25)  = (/1.0000000, 0.99401945, 0.97966272, 0.95182151, 0.90788388, &
       0.84737492, 0.77159661, 0.68326861, 0.58616841, 0.48477158, 0.38389215, 0.28832296,          &
       0.20247590, 0.13002251, 0.073533830, 0.034121160, 0.011142910, 0.0018151600, 7.5820000e-05,  &
       0.0000000, 0.0000000, 0.0000000, 0.0000000, 0.0000000, 0.0000000, 0.0000000/)
 

  integer                                      :: status
  integer                                      :: xbeg_in,ybeg_in,jm_in,im_in,lm_in
  real                                         :: dx_in,dy_in
  integer                                      :: region, imr, jmr, lmr
  real,intent(in)                  :: field3d_in(:,:,:)
  real,intent(inout)               :: field3d_out(:,:,:)

  real, dimension(:,:), pointer     ::  ps

  type(TllGridInfo)                ::  lli_in
  type(TLevelInfo)                 ::  levi_in



  region=1
  lm_in=25
  ps => mass_dat(region)%phlb_t(:,:,1)

  call Init( levi_in, lm_in, a_in, b_in, status )
  im_in=60
  jm_in=45
  xbeg_in=-180
  dx_in=360.0/im_in
  ybeg_in=-90
  dy_in=180.0/jm_in
  call Init( lli_in, xbeg_in+0.5*dx_in, dx_in, im_in, &
                       ybeg_in+0.5*dy_in, dy_in, jm_in, status )

  imr=im(region) ;jmr=jm(region) ;lmr=lm(region)
  call Fill3D( lli(region), levi, 'n', ps(1:imr,1:jmr), field3d_out(1:imr,1:jmr,1:lmr), &
               lli_in, levi_in, field3d_in(:,:,:), 'mass-aver', status )

 end subroutine interp_loss 

   subroutine free_chemloss_ch4_ohcl
    implicit none
    integer               :: region
    do region=1,nregions
     deallocate(kohcl(region)%d3)  ! oh+ch4, cl+ch4 sink
    enddo

   end subroutine free_chemloss_ch4_ohcl
end module chemloss_ch4_ohcl
