
!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################

module emission_ch4_wetl
  !-----------------------------------------------------------------------
  !       purpose
  !       -------
  !       perform ch4 wetl emissions 
  !
  !       EMISSIONS
  !          interface:     call emission_apply_wetl
  !          method:        subroutine is called from its general parent routine apply_emission
  !
  !       COVARIANCE
  !          interface:     call covariance_co2_compute_wetl
  !          method:        subroutine is called from MakePriorUncertainty in enkf_tools.F90
  !
  ! Table of contents:
  !   module emission_ch4_wetl
  !     subroutine declare_emission_ch4_wetl
  !     end subroutine declare_emission_ch4_wetl
  !     subroutine calc_emission_ch4_wetl
  !     end subroutine calc_emission_ch4_wetl
  !     subroutine emission_apply_ch4_wetl(region)
  !     end subroutine emission_apply_ch4_wetl
  !     subroutine free_emission_ch4_wetl
  !     end subroutine free_emission_ch4_wetl
  !   end module emission_ch4_wetl
  !
  !-----------------------------------------------------------------------

  use GO,                  only: gol, goErr, goPr, goLabel
  use Dims,                only: nregions, nlon360,nlat180
  use emission_data,       only: gridbox_per_sec_to_flux
  use global_types,        only: emis_data
  use partools,            only: nmembersloc
  use enkf_common,         only: scalefac
  use emission_common,     only: enkf_is_forward, enkf_is_inverse, EmisInputDir
  use emission_common,     only: ch4_wetl, ch4_wetl_m, ch4_wetl_n, flux_means_are_priors
  use emission_data,       only: do_add_2d, flux_to_gridbox_per_month
  use enkf_constants,      only: nparams_in_state,npar_ocn,npar_land,nreg_ocn,nreg_land

  implicit none

  ! --- const -------------------------
  character(len=*), parameter  ::  mname = 'emission_ch4_wetl'

  private

  ! public routines:
  public   :: declare_emission_ch4_wetl, free_emission_ch4_wetl, calc_emission_ch4_wetl, emission_apply_ch4_wetl

  ! ch4_wetl_regxmem is the result of the calculate routine.
  ! It holds the coarsened flux per ensemble member and region.
  type(emis_data),allocatable,dimension(:,:),target       :: ch4_wetl_regxmem

contains

  subroutine declare_emission_ch4_wetl

    use dims,       only: im,jm,nregions

    implicit none
    integer                            :: status=0
    integer                            :: region, imr, jmr, n

    character(len=*), parameter        :: rname = mname//'/declare_emission_ch4_wetl'

    if(enkf_is_forward()) then
!       print *, "DEBUG: declare_emission_ch4_wetl (forward)",nregions,nmembersloc
       allocate(ch4_wetl_regxmem(nregions, nmembersloc))

       do region=1,nregions
          imr=im(region)
          jmr=jm(region)
!          print *, "DEBUG: ",region, imr,jmr
          do n=1,nmembersloc
             allocate(ch4_wetl_regxmem(region,n)%surf(imr,jmr))
          end do
       enddo
    end if
  end subroutine declare_emission_ch4_wetl

  subroutine calc_emission_ch4_wetl

    use dims,          only: nlon360, nlat180, im,jm, sec_month, newsrun, idate, newmonth, sec_year,newday
    use io_save,       only: ReadFromFile,readtm3hdf
    use emission_data, only: chardate
    use chem_param,    only: xmch4,ra
    use global_data, only : inputdir_ch4
    use toolbox,       only: coarsen_emission_2d, escape_tm
    use go_string,     only: goNum2str

    implicit none
    integer                            :: status=0
    integer                            :: region, imr, jmr,lmr,irec
    integer, parameter                 :: annual_field = 0, rank2 = 2, level1 = 1
    character(len=10)                  :: cdate
    real                               :: year2month
    character (len=75)                           :: filename
    character (len=76)                           ::  filename2
    character (len=5)                            :: varname
    character (len=8)                            :: recname
    character(len=2),dimension(12),parameter     :: mlab=(/'01','02','03','04','05','06','07','08','09','10','11','12'/)    
    character(len=4),dimension(28),parameter :: ylab=(/'1980','1981','1982','1983','1984','1985','1986','1987'&
                                                      ,'1988','1989','1990','1991','1992','1993','1994','1995',&
                                                       '1996','1997','1998','1999','2000','2001','2002','2003',&
                                                       '2004','2005','2006','2007'/)
    integer,dimension(12),parameter :: ndays=(/31,28,31,30,31,30,31,31,30,31,30,31/)

    character(len=5),parameter                   :: source_lab='S0007'

    real,dimension(nlon360,nlat180)    :: ch4_wetl_mem ! flux for a given ensemble member
    integer, parameter                 :: add_field=0
    integer                            :: mem,nreg
    integer                            :: i1,i2,i3,i4,i5,i,j,k,imon,iyr
    integer,parameter                  :: nwpts=3233,nmonth=12,nyear=13,year0=1995
    integer                            :: ilon(nwpts),ilat(nwpts),ifrin(nwpts)
!    real                               :: t_av(nwpts,nmonth),t_mav(nwpts,nmonth,nyear),p_anom(nwpts,nmonth,nyear)
    real                               :: t_anom(nwpts,nmonth,nyear),tp(nmonth),p_anom(nwpts,nmonth,nyear)
    real                               :: walpha(nwpts,nmonth),wbeta(nwpts,nmonth)
    real,dimension(nlon360,nlat180)    :: ch4_wetl_anom ! anomaly array
    real,dimension(nlon360,nlat180)    :: ch4_wa ! temporary anomaly array
    real,dimension(nlon360,nlat180)    :: ch4_w ! temporary array


    character(len=*), parameter        :: rname = mname//'/calc_emission_ch4_wetl'

!  if (newsrun) then
!! read in regression model parameters
!        open(10,file='/lfs0/projects/co2/lori/CH4_input/coefficients_grid_lag', form='formatted')
!        open(71,file='/lfs0/projects/co2/lori/CH4_input/tot_frin_new.dat',form='formatted')
!!        open(60,file='/misc/lfs0/projects/co2/lori/CH4_input/t_era40.dat',form='formatted')
!!        open(61,file='/misc/lfs0/projects/co2/lori/CH4_input/precip_anomalies.dat',form='formatted')
!
!! read longitude, latitude and fractional inundation of each wetland grid cell
!
!       do i=1,nwpts
!          read(71,*) i1,ilon(i),ilat(i),ifrin(i)
!!          print *,i,i1,ilon(i),ilat(i),ifrin(i)
!       enddo
!       close(71)
!
!!read monthly mean temp (oC) and prec (mm/d) anomaly input
!!        do i=1,nwpts
!!          do k=1,nyear
!!             do j=1,nmonth
!!                read(60,*) i1,i2,i3,i4,i5,t_av(i,j),t_mav(i,j,k)
!!!                print *,t_av(i,j),t_mav(i),i,j
!!                read(61,'(4i8,f10.6)') i1,i2,i3,i4,p_anom(i,j,k)
!!!                print *,i,j,k,i1,i2,i3,i4,p_anom(i,j,k)
!!             enddo
!!          enddo
!!       enddo
!!       close(60)
!!       close(61)       
!
!      do k=1,nyear
!         filename2=trim(inputdir_ch4)//'giss_tanom_'//trim(ylab(k+15))//'.dat' 
!!         print *,filename2         
!         open(60,file=filename2,form='formatted')         
!         filename2=trim(inputdir_ch4)//'gpcc_panom_'//trim(ylab(k+15))//'.dat'         
!!         print *,filename2         
!         open(61,file=filename2,form='formatted')
!         do i=1,nwpts           
!          read(60,'(12f10.2)') (t_anom(i,j,k),j=1,nmonth)
!!          print *,t_anom(i,:,k),i,k,'t anomalies'! missing values treated as no temp. anomaly
!          do j=1,nmonth
!           if (abs(t_anom(i,j,k)).ge.1000.) t_anom(i,j,k)=0.
!          enddo           
!          read(61,'(12f10.2)') (tp(j),j=1,nmonth)
!!          print *,tp(:),i,k,'p anomalies'
!          do j=1,nmonth
!! missing values treated as no precip anomaly            
!           if (abs(tp(j)).gt.1000) tp(j)=0.0
!! convert to mm/day from mm/month           
!            p_anom(i,j,k)=tp(j)/ndays(j)           
!          enddo
!!           print *,i,k,p_anom(i,:,k),'precip anomalies'
!         enddo
!         close(60)
!         close(61)
!       enddo
!
!
!!read coefficients
!       do i=1,nwpts
!          do j=1,nmonth
!             read(10,'(i6,i4,2f12.4)') i1,i2,walpha(i,j),wbeta(i,j)
!!             print *,i1,i2,walpha(i,j),wbeta(i,j),i,j
!          enddo
!       enddo
!       close(10)
!
!  endif !newsrun
!  if (newmonth) then
!   ch4_wetl_anom=0.
!
!!calculate CH4 flux anomaly using alpha & beta (mg/(m^2*d))
!
!      do i=1,nwpts
!!       ch4_wetl_anom(ilon(i),ilat(i))=walpha(i,idate(2))*&
!!        (t_mav(i,idate(2),idate(1)-year0+1)-t_av(i,idate(2)))+wbeta(i,idate(2))*p_anom(i,idate(2),idate(1)-year0+1)
!!       ch4_wetl_anom(ilon(i),ilat(i))=ch4_wetl_anom(ilon(i),ilat(i))*ifrin(i)/100.
!!       print*,k,i,ch4_wetl_anom(ilon(i),ilat(i)),ilon(i),ilat(i),idate(2),idate(1)-year0
!! use lagged value of precip anomaly
!       imon=idate(2)-1
!       iyr=idate(1)
!       if (idate(2).eq.1) then
!         imon=12
!         iyr=idate(1)-1
!       endif
!       if (idate(2).eq.1.and.idate(1).eq.1995) then
!         imon=1
!         iyr=idate(1)
!       endif
!       ch4_wetl_anom(ilon(i),ilat(i))=walpha(i,idate(2))*t_anom(i,idate(2),idate(1)-year0+1)+&
!                                      wbeta(i,idate(2))*p_anom(i,imon,iyr-year0+1)
!       ch4_wetl_anom(ilon(i),ilat(i))=ch4_wetl_anom(ilon(i),ilat(i))*ifrin(i)/100.
!!       print*,k,i,ch4_wetl_anom(ilon(i),ilat(i)),ilon(i),ilat(i),idate(2),idate(1)-year0
!
!      enddo
!
!! convert to kg/m2/s
!
!  ch4_wetl_anom=ch4_wetl_anom*1.e-6/86400.
!
!  endif   !newmonth



  if(newday) then
      filename=trim(inputdir_ch4)//'day_CH4_'//source_lab//'_'//mlab(idate(2))//'.hdf'
      varname=source_lab
      irec=idate(3)-1
      if (idate(2).eq.2.AND.idate(3).gt.28) irec=28-1
!      print *,filename,'  ',varname,'  ',irec
      call readtm3hdf(filename,rank2,nlon360,nlat180,level1,irec,ch4_wetl,varname)  ! incoming units are kgCH4/gridbox/sec

       if (status /= 0) call escape_tm('stop in '//rname)
       call gridbox_per_sec_to_flux(ch4_wetl)   ! to kg/m2/s
!       ch4_wetl=(ch4_wetl_anom+ch4_wetl)*1.e3/xmch4  ! to mol/m2/s, add on flux anomaly too
       ch4_wetl=(ch4_wetl)*1.e3/xmch4  ! to mol/m2/s
!       print *,'wetland fluxes::::::::::::',sum(ch4_wetl),sum(ch4_wetl_anom)
       
       ! accumulate weekly average
       !
       if(flux_means_are_priors .eq. .TRUE.) then

       if(allocated(ch4_wetl_m)) then
          ch4_wetl_m=ch4_wetl_m+ch4_wetl ! mol/m2/s
          ch4_wetl_n=ch4_wetl_n+1
       endif

       ! The rest only useful to the forward code
       if(enkf_is_inverse()) return

       ! Apply net flux scaling by ensemble member, change units, and
       ! coarsen to the regions.

       do mem=1,nmembersloc
!          ch4_wetl_mem=ch4_wetl ! mol/m2/s (this is where parameters would be applied for optimized fluxes)
        ch4_wetl_mem=0.0
        do nreg=1,nreg_land
         do i=1,nlon360
          do j=1,nlat180
!           if (scalefac(7+(nreg-1)*10,i,j,mem).ne.0.0) then
            ch4_wetl_mem(i,j)=ch4_wetl_mem(i,j)+ch4_wetl(i,j)*scalefac(7+(nreg-1)*10,i,j,mem)
!           endif
          enddo
         enddo
        enddo

          ! accumulate weekly average
          if(flux_means_are_priors .eq. .FALSE.) then
             ch4_wetl_m=ch4_wetl_m+ch4_wetl_mem ! mol/m2/s
             ch4_wetl_n=ch4_wetl_n+1
          endif
          ch4_wetl_mem=ch4_wetl_mem*xmch4/1.e3 ! from mol/m2/s to kgC/m2/s
          call flux_to_gridbox_per_month(ch4_wetl_mem) ! to kgC/area/month
          call coarsen_emission_2d('ch4_wetl',360,180,ch4_wetl_mem,ch4_wetl_regxmem(:,mem),add_field)
       enddo

    endif
  endif

  end subroutine calc_emission_ch4_wetl

  subroutine emission_apply_ch4_wetl(region)

    ! This routine should only be called by the forward enkf code.  It
    ! applies the wetl ch4 flux by member and region to the iwetl
    ! tracer.  In the inverse code, the calculated flux from each
    ! module is applied to the (single) ch4 tracer.

    use dims,          only: okdebug
    use ParTools,      only: tracer_active
    use chem_param, only         : iwetl, xmc, xmch4, ra


    implicit none

    integer, intent(in)        :: region
    integer                    :: mem

    character(len=*), parameter        :: rname = mname//'/emission_apply_ch4_wetl'

    if(enkf_is_forward()) then
#ifdef MPI
#else
       if(iwetl .gt. 0) then
          do mem=1,nmembersloc
             call do_add_2d(region,iwetl,1,ch4_wetl_regxmem(region,mem)%surf,ra(iwetl),ra(iwetl))
          enddo
       endif
#endif
    endif

    if(okdebug) write(*,*) 'end of emission_apply_ch4_wetl'

  end subroutine emission_apply_ch4_wetl

  subroutine free_emission_ch4_wetl
    use dims,        only: nregions

    implicit none
    integer               :: region,n

    character(len=*), parameter        :: rname = mname//'/free_emission_ch4_wetl'

    if(enkf_is_forward()) then
       do region=1,nregions
          do n=1,nmembersloc
             deallocate(ch4_wetl_regxmem(region,n)%surf)
          end do
       enddo

       deallocate(ch4_wetl_regxmem)
    end if

  end subroutine free_emission_ch4_wetl

end module emission_ch4_wetl
