!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
module emission_common

  ! While the averagers could be put into each flux module, managing
  ! these routines is easier when all the repetitive code is gathered
  ! here in one place.

  use GO,                  only: gol, goErr, goPr, goLabel
  use Dims,                only: nlon360,nlat180,nlon60,nlat45,lm

  public   :: ch4_ocean, ch4_bburn, ch4_coal, ch4_oilgas, ch4_entferm, ch4_rice, ch4_term, &
              ch4_waste, ch4_wildan, ch4_wetl, ch4_soil, ch4_all,ivlon,ivlat,nlat,nlon
  public   :: ch4_ocean_m, ch4_bburn_m, ch4_coal_m, ch4_oilgas_m, ch4_entferm_m, ch4_rice_m, &
              ch4_term_m, ch4_waste_m, ch4_wildan_m, ch4_wetl_m, ch4_soil_m, ch4_all_m
  public   :: ch4_ocean_n, ch4_bburn_n, ch4_coal_n, ch4_oilgas_n, ch4_entferm_n, ch4_rice_n, &
              ch4_term_n, ch4_waste_n, ch4_wildan_n, ch4_wetl_n, ch4_soil_n, ch4_all_n
  public   :: declare_emission_averagers,free_emission_averagers,reset_emission_averagers, calc_emission_averagers, save_emission_averagers
  public   :: declare_emission_ch4, free_emission_ch4
  public   :: enkf_is_inverse, enkf_is_forward, enkf_set_inverse, enkf_set_forward
  public   :: flux_means_are_priors, EmisInputDir

  real,dimension(nlon360,nlat180),target :: ch4_coal
  real,dimension(nlon360,nlat180),target :: ch4_oilgas
  real,dimension(nlon360,nlat180),target :: ch4_entferm
  real,dimension(nlon360,nlat180),target :: ch4_rice
  real,dimension(nlon360,nlat180),target :: ch4_waste
  real,dimension(nlon360,nlat180),target :: ch4_wetl
  real,dimension(nlon360,nlat180),target :: ch4_wildan
  real,dimension(nlon360,nlat180),target :: ch4_term
  real,dimension(nlon360,nlat180),target :: ch4_soil
  real,dimension(nlon360,nlat180),target :: ch4_ocean
  real,dimension(nlon360,nlat180),target :: ch4_bburn
  real,dimension(nlon360,nlat180),target :: ch4_all
  integer, parameter                           :: nlat=54,nlon=94
  integer,dimension(nlon360,nlat180),target :: ivlon(nlon)
  integer,dimension(nlon360,nlat180),target :: ivlat(nlat)

  ! weekly average flux accumulators.  "_m" is the flux field itself,
  ! "_n" is the number of accumulations.

  real,allocatable,dimension(:,:) :: ch4_coal_m
  real,allocatable,dimension(:,:) :: ch4_oilgas_m
  real,allocatable,dimension(:,:) :: ch4_entferm_m
  real,allocatable,dimension(:,:) :: ch4_rice_m
  real,allocatable,dimension(:,:) :: ch4_waste_m
  real,allocatable,dimension(:,:) :: ch4_wetl_m
  real,allocatable,dimension(:,:) :: ch4_wildan_m
  real,allocatable,dimension(:,:) :: ch4_bburn_m
  real,allocatable,dimension(:,:) :: ch4_soil_m
  real,allocatable,dimension(:,:) :: ch4_ocean_m
  real,allocatable,dimension(:,:) :: ch4_term_m
  real,allocatable,dimension(:,:) :: ch4_all_m
  real,allocatable,dimension(:,:,:) :: ch4_loss_m

  integer                                :: ch4_coal_n
  integer                                :: ch4_oilgas_n
  integer                                :: ch4_entferm_n
  integer                                :: ch4_rice_n
  integer                                :: ch4_waste_n
  integer                                :: ch4_wetl_n
  integer                                :: ch4_wildan_n
  integer                                :: ch4_term_n
  integer                                :: ch4_soil_n
  integer                                :: ch4_ocean_n
  integer                                :: ch4_bburn_n
  integer                                :: ch4_all_n
  integer                                :: ch4_loss_n

  ! Are we running in inverse or forward mode?  Set enkf mode in Emis_Init().
  integer,parameter             :: enkf_mode_unknown = 0
  integer,parameter             :: enkf_mode_inverse = 1
  integer,parameter             :: enkf_mode_forward = 2
  integer                       :: enkf_mode = enkf_mode_unknown

  character(len=256)            :: EmisInputDir

  ! Weekly mean fluxes can either be means of the raw priors, or means
  ! of the optimized fluxes.  Default is priors.  This was the
  ! behavior in previous versions, since one needs to verify what the
  ! flux modules are doing, and it is possible to retrieve the scaled
  ! fluxes in postprocessing.  This switch removes the need for that
  ! postprocessing, and may be useful in a future with a more complex
  ! parameter model.
  !
  ! flux_means_are_priors = .TRUE.  -> report weekly means of prior fluxes
  !
  ! flux_means_are_priors = .FALSE.  -> report weekly means of scaled/parameterized/optimized fluxes
  !
  logical                        :: flux_means_are_priors = .TRUE.


contains

    logical function enkf_is_inverse()
      enkf_is_inverse = .FALSE.
      if(enkf_mode .EQ. enkf_mode_inverse) then
         enkf_is_inverse = .TRUE.
      endif
    end function enkf_is_inverse

    logical function enkf_is_forward()
      enkf_is_forward = .FALSE.
      if(enkf_mode .EQ. enkf_mode_forward) then
         enkf_is_forward = .TRUE.
      endif
    end function enkf_is_forward

    subroutine enkf_set_inverse
      enkf_mode=enkf_mode_inverse
      ! print *,'EnKF mode set to inverse'
    end subroutine enkf_set_inverse

    subroutine enkf_set_forward
      enkf_mode=enkf_mode_forward
      ! print *,'EnKF mode set to forward'
    end subroutine enkf_set_forward

  subroutine declare_emission_averagers
!    print *,"DEBUG: in declare_emission_averagers"

!WP! Always write component fluxes to savestate.hdf for analysis
    if(enkf_is_forward()) then
       allocate(ch4_coal_m(nlon360,nlat180))
       allocate(ch4_oilgas_m(nlon360,nlat180))
       allocate(ch4_bburn_m(nlon360,nlat180))
       allocate(ch4_rice_m(nlon360,nlat180))
       allocate(ch4_waste_m(nlon360,nlat180))
       allocate(ch4_wildan_m(nlon360,nlat180))
       allocate(ch4_wetl_m(nlon360,nlat180))
       allocate(ch4_soil_m(nlon360,nlat180))
       allocate(ch4_term_m(nlon360,nlat180))
       allocate(ch4_ocean_m(nlon360,nlat180))
       allocate(ch4_entferm_m(nlon360,nlat180))
       allocate(ch4_loss_m(nlon60,nlat45,lm(1)))
    else
       allocate(ch4_all_m(nlon360,nlat180))
       allocate(ch4_loss_m(nlon60,nlat45,lm(1)))
       allocate(ch4_coal_m(nlon360,nlat180))
       allocate(ch4_oilgas_m(nlon360,nlat180))
       allocate(ch4_bburn_m(nlon360,nlat180))
       allocate(ch4_rice_m(nlon360,nlat180))
       allocate(ch4_waste_m(nlon360,nlat180))
       allocate(ch4_wildan_m(nlon360,nlat180))
       allocate(ch4_wetl_m(nlon360,nlat180))
       allocate(ch4_soil_m(nlon360,nlat180))
       allocate(ch4_term_m(nlon360,nlat180))
       allocate(ch4_ocean_m(nlon360,nlat180))
       allocate(ch4_entferm_m(nlon360,nlat180))
    endif

    call reset_emission_averagers

  end subroutine declare_emission_averagers

  subroutine free_emission_averagers

    if(allocated(ch4_all_m)) then
       deallocate(ch4_all_m)
    endif

    if(allocated(ch4_ocean_m)) then
       deallocate(ch4_ocean_m)
    endif

    if(allocated(ch4_soil_m)) then
       deallocate(ch4_soil_m)
    endif

    if(allocated(ch4_term_m)) then
       deallocate(ch4_term_m)
    endif

    if(allocated(ch4_wildan_m)) then
       deallocate(ch4_wildan_m)
    endif

    if(allocated(ch4_bburn_m)) then
       deallocate(ch4_bburn_m)
    endif

    if(allocated(ch4_waste_m)) then
       deallocate(ch4_waste_m)
    endif

    if(allocated(ch4_wetl_m)) then
       deallocate(ch4_wetl_m)
    endif

    if(allocated(ch4_rice_m)) then
       deallocate(ch4_rice_m)
    endif

    if(allocated(ch4_oilgas_m)) then
       deallocate(ch4_oilgas_m)
    endif

    if(allocated(ch4_coal_m)) then
       deallocate(ch4_coal_m)
    endif

    if(allocated(ch4_entferm_m)) then
       deallocate(ch4_entferm_m)
    endif

    if(allocated(ch4_loss_m)) then
       deallocate(ch4_loss_m)
    endif

  end subroutine free_emission_averagers

  subroutine calc_emission_averagers

    if(allocated(ch4_all_m)) then       
       if(ch4_all_n .GT. 0) then
         ch4_all_m=ch4_all_m/ch4_all_n
      endif
    endif

    if(allocated(ch4_loss_m)) then       
       if(ch4_loss_n .GT. 0) then
         ch4_loss_m=ch4_loss_m/ch4_loss_n
      endif
    endif

    if(allocated(ch4_bburn_m)) then       
       if(ch4_bburn_n .GT. 0) then
         ch4_bburn_m=ch4_bburn_m/ch4_bburn_n
      endif
    endif

    if(allocated(ch4_coal_m)) then       
       if(ch4_coal_n .GT. 0) then
         ch4_coal_m=ch4_coal_m/ch4_coal_n
      endif
    endif

    if(allocated(ch4_oilgas_m)) then       
       if(ch4_oilgas_n .GT. 0) then
         ch4_oilgas_m=ch4_oilgas_m/ch4_oilgas_n
      endif
    endif

    if(allocated(ch4_rice_m)) then       
       if(ch4_rice_n .GT. 0) then
         ch4_rice_m=ch4_rice_m/ch4_rice_n
      endif
    endif

    if(allocated(ch4_entferm_m)) then       
       if(ch4_entferm_n .GT. 0) then
         ch4_entferm_m=ch4_entferm_m/ch4_entferm_n
      endif
    endif

    if(allocated(ch4_wetl_m)) then       
       if(ch4_wetl_n .GT. 0) then
         ch4_wetl_m=ch4_wetl_m/ch4_wetl_n
      endif
    endif

    if(allocated(ch4_waste_m)) then       
       if(ch4_waste_n .GT. 0) then
         ch4_waste_m=ch4_waste_m/ch4_waste_n
      endif
    endif

    if(allocated(ch4_wildan_m)) then       
       if(ch4_wildan_n .GT. 0) then
         ch4_wildan_m=ch4_wildan_m/ch4_wildan_n
      endif
    endif

    if(allocated(ch4_term_m)) then       
       if(ch4_term_n .GT. 0) then
         ch4_term_m=ch4_term_m/ch4_term_n
      endif
    endif

    if(allocated(ch4_soil_m)) then       
       if(ch4_soil_n .GT. 0) then
         ch4_soil_m=ch4_soil_m/ch4_soil_n
      endif
    endif

    if(allocated(ch4_ocean_m)) then       
       if(ch4_ocean_n .GT. 0) then
         ch4_ocean_m=ch4_ocean_m/ch4_ocean_n
      endif
    endif


  end subroutine calc_emission_averagers

  subroutine save_emission_averagers(hdf_file)

    use io_hdf
    use file_hdf,             only: WriteAttribute, THdfFile
    use ParTools,             only: myid, root

    type(THdfFile), intent(in)         ::  hdf_file

    integer               :: status
    character(len=200)    :: dsname,attname,comp,pp
    
    if(myid .ne. root) return
    pp='post'
    if(flux_means_are_priors .eq. .True.) then
       pp='prior'
    endif

    if(allocated(ch4_all_m)) then
       comp='all'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       print *,'save_emission_averagers:',dsname,attname,nlon360,nlat180
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_all_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_all_n, status)
       print *,'save_emission_averagers:'
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_loss_m)) then
       comp='loss'
       dsname='loss_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='loss_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon60,'lon',nlat45,'lat',lm(1),'lev',ch4_loss_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_loss_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean loss '//trim(dsname)
    endif

    if(allocated(ch4_term_m)) then
       comp='term'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_term_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_term_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_coal_m)) then
       comp='coal'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_coal_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_coal_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_oilgas_m)) then
       comp='oilgas'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_oilgas_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_oilgas_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_rice_m)) then
       comp='rice'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_rice_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_rice_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_waste_m)) then
       comp='waste'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_waste_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_waste_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_wetl_m)) then
       comp='wetl'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_wetl_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_wetl_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_wildan_m)) then
       comp='wildan'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_wildan_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_wildan_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_bburn_m)) then
       comp='bburn'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_bburn_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_bburn_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_entferm_m)) then
       comp='entferm'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_entferm_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_entferm_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_soil_m)) then
       comp='soil'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_soil_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_soil_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

    if(allocated(ch4_ocean_m)) then
       comp='ocean'
       dsname='flux_'//trim(comp)//'_'//trim(pp)//'_mean'
       attname='flux_'//trim(comp)//'_'//trim(pp)//'_n'
       call io_write(hdf_file%id,nlon360,'lon',nlat180,'lat',ch4_ocean_m,dsname,units='mol m-2 s-1')
       call WriteAttribute( hdf_file, attname, ch4_ocean_n, status)
       IF_ERROR_RETURN(status=1)
       print *,'[save_emission_averagers] wrote weekly mean flux '//trim(dsname)
    endif

  end subroutine save_emission_averagers

  subroutine reset_emission_averagers

    if(allocated(ch4_all_m)) then
       ch4_all_m=0.
       ch4_all_n=0
    endif

    if(allocated(ch4_loss_m)) then
       ch4_loss_m=0.
       ch4_loss_n=0
    endif

    if(allocated(ch4_ocean_m)) then
       ch4_ocean_m=0.
       ch4_ocean_n=0
    endif

    if(allocated(ch4_oilgas_m)) then
       ch4_oilgas_m=0.
       ch4_oilgas_n=0
    endif

    if(allocated(ch4_coal_m)) then
       ch4_coal_m=0.
       ch4_coal_n=0
    endif

    if(allocated(ch4_wetl_m)) then
       ch4_wetl_m=0.
       ch4_wetl_n=0
    endif

    if(allocated(ch4_rice_m)) then
       ch4_rice_m=0.
       ch4_rice_n=0
    endif

    if(allocated(ch4_wildan_m)) then
       ch4_wildan_m=0.
       ch4_wildan_n=0
    endif

    if(allocated(ch4_entferm_m)) then
       ch4_entferm_m=0.
       ch4_entferm_n=0
    endif

    if(allocated(ch4_waste_m)) then
       ch4_waste_m=0.
       ch4_waste_n=0
    endif

    if(allocated(ch4_term_m)) then
       ch4_term_m=0.
       ch4_term_n=0
    endif

    if(allocated(ch4_bburn_m)) then
       ch4_bburn_m=0.
       ch4_bburn_n=0
    endif

    if(allocated(ch4_soil_m)) then
       ch4_soil_m=0.
       ch4_soil_n=0
    endif

  end subroutine reset_emission_averagers

end module emission_common
