module diffusion

  use dims, only                  : nregions, revert, czetak
  use binas, only                 : xmair
  use chem_param, only            : emis_data

  implicit none
  
  ! --- in/out ------------------------------------

  private
  public   ::  kzz_online, kzz_offline, calc_kzz
  
  
  ! --- const ------------------------------------

  logical, parameter :: kzz_offline = .false.            ! use pre-calculated kzz fields
  logical, parameter :: kzz_online  = .not. kzz_offline

  ! --- local ------------------------------------

  type(emis_data),dimension(nregions),target       :: lsmask,sr_ecm,sr_ols
  type(emis_data),dimension(nregions),target       :: u10m,v10m,slhf,sshf,ss, ustar_loc, sr_mix

  !--------------------------------------------------------------------------------
  ! global input fields:
  ! phlb(:,:,lm+1)          ! half level pressure [Pa]
  ! gph(:,:,lm+1)           ! height of half-levels [m]    !CMK note gph starts NOW at 1!
  ! t                       ! temperature [K]
  ! q                       ! specific humidity [kg/kg]
  ! uwind                   ! mean wind W-E [m/s]
  ! vwind                   ! mean wind S-N [m/s]
  ! sshf                    ! surface sensible heat flux [W / m^2]
  ! slhf                    ! surface latent heat flux [W / m^2]
  ! ustar (ustar_loc)          ! velocity scale
  !
  ! global output fields:
  ! kvh                     ! eddy diffusivity for heat at top [m^2/s]
  ! pblh                    ! boundary layer height[m]
  ! dkg                                     ! vertical diffusion coefficient [ks s-1]
  ! Adjusted for TM5 by Bram Bregman
  ! April 2005/ adjoint version
  !--------------------------------------------------------------------------------


contains

  !------------------------------------
  !
  ! Purpose:
  ! -------
  ! this subroutine reads and prepares all fields needed for the calculation of kzz
  !
  ! External
  ! --------
  ! dd_get_3_hourly_surface_fields
  ! dd_calc_ustar_raero_rb
  ! dd_coarsen_fields
  !
  ! Reference
  ! ---------
  ! Ganzeveld and Lelieveld (1996) and references therein.
  ! Adjusted for TM5, Bram Bregman, August 2003
  !
  !

  subroutine calc_kzz

    use binas,       only: vkarman, grav
    use dims,        only: im, jm, lmax_conv, idate, lm, okdebug
    use dims,        only: nread, ndyn, tref, at, bt
    use dims,        only: isr,jsr,ier,jer
    use global_data, only: meteo_dat, mass_dat, conv_dat, wind_dat, region_dat
    use ParTools,   only: myid, root

    ! --- in/out ----------------------------

    ! --- local ------------------------------

    character(len=10)                            :: c_time
    real,dimension(:,:,:),pointer                :: phlb,gph,t,q,m,pu,pv
    real,dimension(:,:,:),pointer                :: am,bm,cm ! fluxes in kg/timestep(region)
    real,dimension(:,:,:),pointer                :: dkg     !vert diff (ks/s)
    real,dimension(:),pointer                    :: dxyp
    integer,parameter                            :: avg_field=1,nlon360=360,nlat180=180
    integer                                      :: i,region, nsend, ntimes
    real, dimension(:,:,:),allocatable,target    :: m_adj, phlb_adj
    real, dimension(:,:),allocatable             :: p_adj
    integer                                      :: imr, jmr, lmr, j,l

    ! --- begin ----------------------------------

    if (myid.eq.root.and.okdebug) write(*,*) 'start of calc_kzz'

!#ifdef MPI
!    if(myid == root_k) then    ! reading /calculations by root!
!#endif

      do region = 1, nregions
        allocate(u10m(region)%surf(im(region),jm(region)))
        allocate(v10m(region)%surf(im(region),jm(region)))
        allocate(slhf(region)%surf(im(region),jm(region)))
        allocate(sshf(region)%surf(im(region),jm(region)))
        allocate(ss(region)%surf(im(region),jm(region)))
        allocate(ustar_loc(region)%surf(im(region),jm(region)))
        allocate(sr_ols(region)%surf(im(region),jm(region)))
        allocate(sr_ecm(region)%surf(im(region),jm(region)))
        allocate(sr_mix(region)%surf(im(region),jm(region)))
        allocate(lsmask(region)%surf(im(region),jm(region)))
     end do

      ! -- for output of time header on debug fields

      write(c_time,'(i4,3(i2))') idate(1:4)   !CMK corrected

      do i=1,10 
        if (c_time(i:i)==' ') c_time(i:i)='0'
      end do

      !
      ! surface data sets may have the following characteristics
      ! 1. instanteneous. The time written in the file is valid from t-1.5 to t+1.5
      ! 2. Accumulated. The time written in the file is valid from t to t+3
      ! 3. Daily averaged. The time on the file is always 00 hours, and valid until t+24
      !
      ! ***   It is essential to understand this timing when reading and applying these sets.
      !
      !
      ! fd2mk it has to be decided to 'save' fields like vgrat_low and daily average surface fields
      ! for later use, or to read it every 3 hours time step

      ! lsm
      call dd_get_land_sea_mask                                                                       

      ! sr_ecm,sr_ols
      call dd_get_daily_av_surf_fields                                                        

      ! wind10m,slhf,sshf
      call dd_get_3_hourly_surface_fields                         

      do region = 1,nregions

        if (revert == -1) then 
           !mkadj
           !mass and pressure are at t=t+dt in adjoint
           !these should be brought back to t to get
           !exactly the same diffusion as in forward run
           imr = im(region) ; jmr = jm(region) ; lmr = lm(region)
           allocate ( m_adj(1:imr,1:jmr,1:lmr) )
           allocate ( phlb_adj(imr,jmr,lmr+1))
           allocate ( p_adj(imr,jmr))
           am => wind_dat(region)%am_t
           bm => wind_dat(region)%bm_t
           cm => wind_dat(region)%cm_t
           dxyp => region_dat(region)%dxyp
           ntimes = (nread/(ndyn))*tref(region)
           m_adj = mass_dat(region)%m_t(1:imr,1:jmr,1:lmr)
           do i=1,ntimes
              m_adj = m_adj + revert*am(0:imr-1,1:jmr  ,1:lmr  ) &   !east   
                            - revert*am(1:imr  ,1:jmr  ,1:lmr  ) &   !west
                            + revert*bm(1:imr  ,1:jmr  ,1:lmr  ) &   !south
                            - revert*bm(1:imr  ,2:jmr+1,1:lmr  ) &   !north
                            + revert*cm(1:imr  ,1:jmr  ,0:lmr-1) &   !lower
                            - revert*cm(1:imr  ,1:jmr  ,1:lmr  )     !upper
           enddo
           p_adj(:,:) = 0.0
           do l=1,lmr
              ! note that on the EDGES masses are coarse>
              ! diffusion iis applied only on core zoom>
              do j=jsr(region), jer(region)  
                 do i=isr(region), ier(region)
                    p_adj(i,j) = p_adj(i,j) + m_adj(i,j,l)*grav/dxyp(j)
                 end do
              end do
           end do
           do l=1,lmr+1
              do j=jsr(region), jer(region)  
                 do i=isr(region), ier(region)
                    phlb_adj(i,j,l) = at(l)+bt(l)*p_adj(i,j)
                 end do
              end do
           end do
           deallocate(p_adj)
           nullify(am)
           nullify(bm)
           nullify(cm)
           nullify(dxyp)
           
           phlb => phlb_adj
           m => m_adj
        else
           phlb => mass_dat(region)%phlb_t
           m => mass_dat(region)%m_t
        endif
          
        pu => wind_dat(region)%pu_t
        pv => wind_dat(region)%pv_t
        gph => meteo_dat(region)%gph
        t => meteo_dat(region)%t
        q => meteo_dat(region)%q
        dkg => conv_dat(region)%dkg

        ! ustar_loc
        call dd_calc_ustar(region)

        ! calculate kzz                                                         
        call bldiff(region)                                                                                     

        nullify(phlb)
        nullify(m)
        nullify(pu)
        nullify(pv)
        nullify(dkg)
        nullify(t)
        nullify(q)
        nullify(gph)
        if(revert == -1) then
           deallocate(m_adj)
           deallocate(phlb_adj)
        endif
      end do

!#ifdef MPI
!    end if   !myid == root_k
!#endif

!#ifdef MPI
!    do region = 1, nregions
!      nsend = im(region)*jm(region)*lmax_conv 
!      if(lmloc.ne.0) call mpi_bcast(conv_dat(region)%dkg ,nsend, my_real, root_k, com_lev, ierr)
!    end do
!
!    if ( myid == root_k ) then
!#endif
!      if (okdebug) print *, 'vertical diffusion kzz (dkg) broadcasted' 

      do region = 1,nregions
        deallocate(slhf(region)%surf)
        deallocate(sshf(region)%surf)
        deallocate(ss(region)%surf)
        deallocate(ustar_loc(region)%surf)
        deallocate(sr_ecm(region)%surf)
        deallocate(sr_ols(region)%surf)
        deallocate(u10m(region)%surf)
        deallocate(v10m(region)%surf)
        deallocate(sr_mix(region)%surf)
        deallocate(lsmask(region)%surf)
      end do

!#ifdef MPI
!    endif
!#endif

    if (myid.eq.root.and.okdebug) write(*,*) 'end of calc_kzz'

  contains

    !
    ! routine to read fields that need to be initialised only once
    !

    !-----------------------------------
    ! Purpose
    ! -------
    ! read ECMWF dataset for landmask
    ! 
    ! External
    ! --------
    ! read_meteo
    ! dd_field_statistics and dd_field_statistics_i8
    !
    ! Reference
    ! ----------
    ! None
    !------------------------

    subroutine dd_get_land_sea_mask

      use MeteoFiles, only: ReadSurfaceField
      use chem_param, only: emis_data
      use datetime,   only: date2tau, tau2date
      use toolbox,    only: escape_tm, coarsen_emission
      
      ! --- local ------------------------------------

      integer,dimension(6)                         :: idater_acc,idater

      integer                                      :: i,j,nv,itaux,region
      integer(kind=1),dimension(:,:), allocatable  :: field                     
      real,dimension(:,:), allocatable             :: field_in                     
      integer,parameter                            :: avg_field=1,nlon360=360,nlat180=180

      integer            ::  status

      allocate(field(nlon360,nlat180))
      allocate(field_in(nlon360,nlat180))

      call date2tau(idate,itaux)
      if (revert == 1) then
         call tau2date(itaux-3600*21,idater)!BB time shift needed since for hh ==21 next day file
      else   
         call tau2date(itaux-3600*24,idater)!BB time shift needed since for hh ==21 next day file 
      endif                                                                             ! is read in, but day in file is one day less
      idater(4) = 21  !BB in file daily data at fixed hour

      call ReadSurfaceField( field_in, 'lsm', idater, 24, status )
      if (status/=0) call escape_tm( 'dd_get_land_sea_mask: error reading lsm' )

      ! coarsen the field
      call coarsen_emission('lsm',nlon360,nlat180,field_in,lsmask,avg_field)

      deallocate(field)
      deallocate(field_in)

      if (okdebug) write(*,*) 'end of dd_get_land_sea_mask'

    end subroutine dd_get_land_sea_mask
    
    
    ! ***

    !---------------------------------------------
    !
    ! Purpose
    ! -------
    ! read daily averaged ECMWF datasets
    ! 
    ! External
    ! --------
    ! read_meteo
    ! dd_field_statistics and dd_field_statistics_i8
    !
    ! Reference
    ! ----------
    ! None
    !------------------------

    subroutine dd_get_daily_av_surf_fields

      use MeteoFiles, only: ReadSurfaceField
      use datetime,   only: date2tau, tau2date
      use toolbox,    only: escape_tm, coarsen_emission

      ! --- local ---------------------------------

      integer                            :: itaux
      integer,dimension(6)               :: idater
      real,dimension(:,:), allocatable   :: field                     
      integer,parameter                  :: avg_field=1,nlon360=360,nlat180=180

      integer            ::  status

      ! --- begin ----------------------------------
      
      allocate(field(nlon360,nlat180))

      call date2tau(idate,itaux)

      if (revert == 1) then
         call tau2date(itaux-3600*21,idater)  !BB time shift needed since for hh ==21 next day file
      else
         call tau2date(itaux-3600*24,idater)  !BB time shift needed since for hh ==21 next day file
      endif           !is read in, but day in file is one day less
      idater(4) = 21  !BB in file daily data at fixed time

      call ReadSurfaceField( field, 'sr_ecm', idater, 24, status )
      if (status/=0) call escape_tm( 'dd_get_daily_av_surf_fields: error reading sr_ecm' )

      if (okdebug) call dd_field_statistics('sr_ecm',field,nlon360,nlat180)

      call coarsen_emission('sr_ecm',nlon360,nlat180,field,sr_ecm,avg_field)

      call ReadSurfaceField( field, 'sr_ols', idater, 24, status )
      if (status/=0) call escape_tm( 'dd_get_daily_av_surf_fields: error reading sr_ecm' )

      if (okdebug) call dd_field_statistics('sr_ols',field,nlon360,nlat180)
      !if (okdebug) call dumpfield(0,'sr_ols.hdf',field,'sr_ols')

      ! coarsen the field
      call coarsen_emission('sr_ols',nlon360,nlat180,field,sr_ols,avg_field)

      deallocate(field)

    end subroutine  dd_get_daily_av_surf_fields
    
    
    ! ***

    !----------------------------------------
    !
    ! Purpose
    ! -------
    ! read 3-hourly ECMWF datasets, can be instanteneous or acccumulated
    ! 
    ! External
    ! --------
    ! date2tau
    ! tau2date
    ! read_meteo
    ! dd_field_statistics and dd_field_statistics_i8
    !
    ! Reference
    ! ----------
    ! None
    !------------------------

    subroutine dd_get_3_hourly_surface_fields

      use MeteoFiles, only: ReadSurfaceField
      use datetime,   only: date2tau, tau2date
      use toolbox,    only: escape_tm, coarsen_emission

      ! --- const ------------------------------------------
                           
      integer,parameter                    :: avg_field=1,nlon360=360,nlat180=180
      
      ! --- local --------------------------------------------
      
      integer,dimension(6)                 :: idater_acc,idater
      integer                              :: itaux ! auxiliary variable
      real,dimension(:,:), allocatable     :: field
      
      integer            ::  status

      ! --- begin ---------------------------------------------

      allocate(field(nlon360,nlat180)) 
      
      !
      !     [t   ,        t+3]
      !
      !    idate
      !    idater_acc     idater
      !

      call date2tau(idate,itaux)
      call tau2date(itaux,idater_acc)
      if (revert == 1) then
         call tau2date(itaux+3*3600,idater) !CMK use 00 for 21-00, 03 for 00-03 etc... (more handy!) 
         call tau2date(itaux,idater_acc)      
      else
         call tau2date(itaux,idater)
         call tau2date(itaux-3*3600,idater_acc)        !CMKADJ
      endif
      if (okdebug) write(*,*) 'dd_get_3_hourly_surface_fields read time on instanteneous field',idater
      if (okdebug) write(*,*) 'dd_get_3_hourly_surface_fields read time on accumulated field',idater_acc

      !
      ! read instanteneous fields
      !

      ! u-component [East West] component of diagnosed 10m wind [m/s]
      call ReadSurfaceField( field, 'u10m', idater, 0, status )
      if (status/=0) call escape_tm( 'dd_get_3_hourly_surface_fields: error from ReadSurfaceField' )
      if (okdebug) call dd_field_statistics('u10m',field,nlon360,nlat180)

      call coarsen_emission('u10m',nlon360,nlat180,field,u10m,avg_field)

      ! v-component [North-South] component of diagnosed 10 m wind [m/s]
      call ReadSurfaceField( field, 'v10m', idater, 0, status )
      if (status/=0) call escape_tm( 'dd_get_3_hourly_surface_fields: error from ReadSurfaceField' )
      if (okdebug) call dd_field_statistics('v10m',field,nlon360,nlat180)

      ! coarsen the field
      call coarsen_emission('v10m',nlon360,nlat180,field,v10m,avg_field)
      !
      ! wind at 10 meters is diagnosed from wind at higher levels
      ! documentation can be found on http:www.ecmwf.int/ifsdocs/research/PHYSICS   chapter 3
      !

      !
      ! read accumulated fields
      !

      call ReadSurfaceField( field, 'sshf', idater_acc, 3, status )
      if (status/=0) call escape_tm( 'dd_get_3_hourly_surface_fields: error from ReadSurfaceField' )

      if (okdebug) call dd_field_statistics('sshf',field,nlon360,nlat180)

      ! coarsen the field
      call coarsen_emission('sshf',nlon360,nlat180,field,sshf,avg_field)

      call ReadSurfaceField( field, 'slhf', idater_acc, 3, status )
      if (status/=0) call escape_tm( 'dd_get_3_hourly_surface_fields: error from ReadSurfaceField' )
      if (okdebug) call dd_field_statistics('slhf',field,nlon360,nlat180)

      ! coarsen the field
      call coarsen_emission('slhf',nlon360,nlat180,field,slhf,avg_field)

      call ReadSurfaceField( field, 'sstr', idater_acc, 3, status )
      if (status/=0) call escape_tm( 'dd_get_3_hourly_surface_fields: error from ReadSurfaceField' )

      if (okdebug) call dd_field_statistics('ss',field,nlon360,nlat180)

      call coarsen_emission('sstr',nlon360,nlat180,field,ss,avg_field)

      if (okdebug) write(*,*) 'end of dd_get_3_hourly_surface_fields'

    end subroutine dd_get_3_hourly_surface_fields
    
    
    ! ***


    !--------------------------------
    !
    ! Purpose
    ! -------
    ! Calculate friction velocity (ustar)
    ! Method
    ! ------
    ! uses the log normal wind profile information stored by ECMWF in 10 meter wind
    ! to estimate a local ustar over land
    ! uses Charnock equation over sea to estimate ustar
    ! aerodynamic resistance is calculated from heat fluxes and ustar using a constant reference height
    ! sub laminar resistance from ustar
    ! 
    ! External
    ! --------
    ! dd_field_statistics 
    ! dumpfield
    !
    ! Reference
    ! ----------
    ! Ge Verver (personal communication, 2003)
    !------------------------

    subroutine dd_calc_ustar(region)

      use binas, only : grav, vKarman

      !--- in/out ------------------------------------

      integer,intent(in)              :: region
      
      ! --- const -------------------------------------

      real, parameter :: Href=30.     ! constant reference height for calculations of raero
                                      ! some constants specific to calculation of ustar and raero
      real,parameter  ::alfa_charnock1=0.11,rhoCp=1231.0,rhoLv=3013000.0,alfa_charnock2=0.018,v_charnock=1.5e-5 !(m2/s)
      real, parameter :: rz0 =2.0

      !--- local -------------------------------------

      integer ::i,j

      real :: ustar_sea,ustar_land,xland,sr_sea,sr_land,sr_help,wind10m
      real,dimension(:,:),allocatable :: field

      ! --- begin ----------------------------------

      allocate(field(im(region),jm(region)))

      do j=1,jm(region)
        do i=1,im(region)
          xland=lsmask(region)%surf(i,j)/100.

          ! SEA:
          ! surface roughness  from Charnock equation
          ! friction velocity from surface stress
          !

          ustar_sea=sqrt(ss(region)%surf(i,j))
          sr_sea=alfa_charnock1*v_charnock/ustar_sea + alfa_charnock2*ss(region)%surf(i,j)/grav 

          !
          ! LAND
          ! calculate the 'local' surface roughness for momentum that is consistent with 10 m wind
          ! and the Olsson data base, we assume that the windspeed at 75 m is independent of
          ! surface roughness
          !
          wind10m=sqrt(u10m(region)%surf(i,j)*u10m(region)%surf(i,j)+ &
                                   v10m(region)%surf(i,j)*v10m(region)%surf(i,j)) 

          if (xland>0.) then
            sr_land=max(1e-2,sr_ols(region)%surf(i,j))  ! occurs at Islands, etc.
            sr_help=min(sr_ecm(region)%surf(i,j),0.03)
            !fd ustar_land=vKarman*wind10m(i,j)/alog(10./sr_help) !ustar consistent with 'clipped' large scale roughness
            ustar_land=vKarman*wind10m/&
                                    alog(10./sr_help)*alog(75./sr_help)/alog(10./sr_land)
          else
            sr_land = 0.0
            ustar_land = 0.0
          endif

                  ustar_loc(region)%surf(i,j)=xland*ustar_land+(1-xland)*ustar_sea
                  sr_mix(region)%surf(i,j) =xland*sr_land  +(1-xland)*sr_sea
        enddo !i
      enddo !j
      field = ustar_loc(region)%surf

      if (okdebug) call dd_field_statistics('ustar_loc',field,im(region),jm(region))
      !if (okdebug) call dumpfield(0,'ustar_loc'//c_time//'.hdf',field,'ustar_loc')

      deallocate(field)

      if (okdebug) write(*,*) 'end calc_ustar'
    
    end subroutine dd_calc_ustar
    
    
    ! ***
    

    !------------------------------------------------------------------
    ! Procedure:
    ! (1) calc_Kv: Diffusion coefficients full atmosphere from shear
    ! (2) pblhght: Determine the boundary layer height
    ! (3) difcoef: Replace diffusion coefficients in the boundary layer
    !------------------------------------------------------------------
    ! subroutine adjusted for TM5, Bram Bregman, August 2003.
    !-------------------------IO---------------------------------------

    subroutine bldiff(region)

      use binas,   only: ae, cp_air, Rgas, grav, Lvap
      use dims,    only: gtor, dx, dy, ybeg, xref, yref
      use dims,    only: isr, ier, jsr, jer
      use toolbox, only: dumpfield

      ! --- in/out ----------------------------------------------
      
      integer,intent(in)              :: region

      ! --- const ----------------------------------
      
      ! minimum value diffusion coefficient
      real, parameter          :: ckmin = 1.e-15

      real, parameter          :: sffrac=  0.1
      real, parameter          :: onet   = 1./3.
      real, parameter          :: rair   = Rgas*1.e3/xmair
      real, parameter          :: ccpq   = 1869.46/cp_air -1.
      real, parameter          :: epsilo = rair/461.51
      real, parameter          :: apzero = 100000. ! reference pressure (usually 1000 hPa)
      real, parameter          :: ccon   = sffrac*vkarman
      real, parameter          :: betam=15.,betah=15.
      real, parameter          :: binm   = betam*sffrac
      real, parameter          :: binh   = betah*sffrac
      real, parameter          :: fr_excess = 1.
      real, parameter          :: fak = fr_excess * 8.5
      real, parameter          :: fakn = fr_excess * 7.2

      real, parameter          :: zkappa = rair / cp_air
      real, parameter          :: zcrdq = 1.0/epsilo - 1.0

      real, parameter          :: tiny     = 1.e-9   ! bound wind**2 to avoid dividing by 0
      real, parameter          :: zacb     = 100.    !factor in ustr-shearproduction term
      real, parameter          :: ricr     = 0.3     ! critical richardson number

      real, parameter          :: ssfrac=0.1

      ! --- local ------------------------------------

      real, dimension(:,:,:), allocatable:: wkvf,&     ! outer layer Kv (at the top of each layer)
                                            zdup2,&    ! vertical shear squared
                                            zrinub,&   ! Richardson number
                                            wthm,&     ! potential temperature
                                            zthvk,&    ! virtual potential temperature
                                            pf,&       ! full level pressure
                                            whm, &     ! height of layer centers
                                            uwind,vwind,& ! wind velocities [m s-1]
                                            kvh                     ! 

      real, dimension(:,:), allocatable  :: ustr,&     ! velocity scale     
                                            wheat,&    ! surface sensible heat flux [K m/s]
                                            wqflx,&    ! surface latent heat flux [kg m/ (kg s)]
                                            wobkl,&    ! Monin Obukhov length
                                            wheatv,&   ! surface virtual heat flux [K m/s]
                                            wtseff   ! theta_v at lowest level [K]

      integer             :: i, j, l, intfac

      ! variables for the calculation of free atmosphere vertical diffusion coefficients wkvf
      real                        :: cml2,zlambdac,arg,zdz,zthva,zfunst, zfstabh,zsstab,zkvn,dyy
      real,dimension(jm(region))  :: dxx,lat
      integer                     :: jq! jqif function

      ! variabales for the calculation of the boundary layer height
      real                        :: wrino      ! Richardson number
      real                        :: thvparc    ! parcel virtual potential temperature
      real                        :: fmt,wsc,tkv,zrino,vvk,dtkv,zexcess
      integer                     :: jwork 
      real,dimension(:,:),pointer   :: pblh         ! planetary boundary layer height [m]

      ! variables to calculate difusion coefficient in the boundary layer
      !real                        :: wsc,cml2
      real                        :: z,zwstr,zkvh,kvhmin,zm,zp,zslask,zsl1, &
                                     zh,zzh,zl,pr,zpblk,zfstabh1,zfstabh2,&
                                     term1,term2,term3,obukov
      integer                     :: jq1,jq2,jq3,jq4,jq5,jck,jqq

      real                        :: yres
      real                        :: xres
      
      ! --- begin ------------------------------------------

      if (okdebug) write(*,*) 'start of bldiff'

      allocate(wkvf(im(region),jm(region),lm(region)))
      allocate(zdup2(im(region),jm(region),lm(region)))
      allocate(zrinub(im(region),jm(region),lm(region)))
      allocate(wthm(im(region),jm(region),lm(region)))
      allocate(zthvk(im(region),jm(region),lm(region)))
      allocate(pf(im(region),jm(region),lm(region)))
      allocate(whm(im(region),jm(region),lm(region)))
      allocate(uwind(im(region),jm(region),lm(region)))
      allocate(vwind(im(region),jm(region),lm(region)))
      allocate(kvh(im(region),jm(region),lm(region)))
      allocate(ustr(im(region),jm(region)))
      allocate(wheat(im(region),jm(region)))
      allocate(wqflx(im(region),jm(region)))
      allocate(wobkl(im(region),jm(region)))
      allocate(wheatv(im(region),jm(region)))
      allocate(wtseff(im(region),jm(region)))
      pblh => conv_dat(region)%blh
!ckm      allocate(pblh(im(region),jm(region)))

      ! mid-layer pressure levels
      ! potential temperature field
      ! virtual potential temperature field
      !
      do l=1,lm(region)
         do j=jsr(region), jer(region)
            do i=isr(region), ier(region)
               pf(i,j,l)    = (phlb(i,j,l) + phlb(i,j,l+1) )*0.5
               wthm(i,j,l)  = t(i,j,l) * ( apzero/pf(i,j,l) )**zkappa
               zthvk(i,j,l) = wthm(i,j,l) * (1.0 + zcrdq*q(i,j,l))
            end do
         end do
      end do

      ! mid-layer heights (note: gph array is for levels 1:lm+1) 
      !
      do l=1,lm(region)
         do j=jsr(region), jer(region)
            do i=isr(region), ier(region)
               intfac = (phlb(i,j,l)-pf(i,j,l)) / (phlb(i,j,l)-phlb(i,j,l+1))
               whm(i,j,l) = (gph(i,j,l+1)-gph(i,j,1)) * (1.0-intfac) + (gph(i,j,l+1)  -gph(i,j,1)) * intfac
            end do
         end do
      end do

      ! convert units of sensible and latent heat fluxes
      ! sshf: W/m2 / (J/kg/K) / (kg/m3) =  J/s/m2 /J kg K /kg m3 =  K m/s
      ! slhf: W/m2 / (J/kg) / (kg/m3) = J/s/m2 /J kg /kg m3 = m/s
      ! ustr is local velocity scale,  NOT from ECMWF surface stresses
      !
      do j=jsr(region), jer(region)
         do i=isr(region), ier(region)
            wheat(i,j) = sshf(region)%surf(i,j) * (gph(i,j,2)-gph(i,j,1)) * grav / &
                                  (phlb(i,j,1) - phlb(i,j,2)) / cp_air 
            wqflx(i,j) = slhf(region)%surf(i,j) * (gph(i,j,2)-gph(i,j,1)) * grav / &
                                  (phlb(i,j,1) - phlb(i,j,2)) / Lvap
            ! if from surface stress:   ustr(:,:) = sqrt(sstr(:,:))
            ustr(i,j) = max(ustar_loc(region)%surf(i,j), 0.01)
            wheat(i,j)=-1*wheat(i,j)   !WP! bug fix
            wqflx(i,j)=-1*wqflx(i,j)   !WP! bug fix
         enddo
      enddo
      !
      !-----------------------------------------------------------------
      ! compute the free atmosphere vertical diffusion coefficients wkvf
      ! height-dependent asymptotic mixing length implemented
      ! Holtslag and Boville, 1993, J. Climate, 6, 1825-1842.
      !------------------------------------------------------------------

      ! first calculate wind velocities [m s-1] from the mass fluxes

      yres = dy/yref(region)
      xres = dx/xref(region)

      do j = 1,jm(region)
          lat(j) = ybeg(region) + 0.5 * yres + yres * (j-1)
      enddo

      dxx(:) = ae * xres * gtor * cos(lat(:)*gtor)
      dyy = ae * yres * gtor

      do l=1,lm(region)
        do j=jsr(region), jer(region)
          do i=isr(region), ier(region)
             uwind(i,j,l) = dxx(j)*(pu(i,j,l) + pu(i-1,j,l))*0.5 / m(i,j,l)
             ! cmk Bug repared, 06/2005 yres --> dyy
             vwind(i,j,l) = dyy*  (pv(i,j,l) + pv(i,j+1,l))*0.5 / m(i,j,l)
          enddo
        enddo
      enddo

      !    
      ! Initialize gradient Richardson number and free tropospheric k values
      !

      if (okdebug) write(*,*) '  Richardson number'

      zrinub(:,:,:) = 0.0
      wkvf(:,:,:)   = 0.0
      zdup2(:,:,:)  = 0.0
      !
      do l=1,lm(region)-1
        do j=jsr(region), jer(region)
          do i=isr(region), ier(region)
            !
            ! if (delta gph < 1000) then jq=1
            jq = jqif( 1000.0, (gph(i,j,l+1)-gph(i,j,1)) )
            zlambdac=jq*300.+(1-jq)*(30. + 270. * exp (1.-(gph(i,j,l+1)-gph(i,j,1))/1000.))
            cml2=(1./( 1./zlambdac + 1./(vkarman*(gph(i,j,l+1)-gph(i,j,1))) ))**2.
            !
            !  vertical shear squared,
            !  min value of (delta v)**2 prevents zero shear
            !
            zdup2(i,j,l) = (uwind(i,j,l+1)-uwind(i,j,l))**2 + &
                                               (vwind(i,j,l+1)-vwind(i,j,l))**2
            zdup2(i,j,l) = max(zdup2(i,j,l),1.e-10)
            zdz   = whm(i,j,l+1) - whm(i,j,l)
            zdup2(i,j,l) = zdup2(i,j,l)/(zdz**2)
            !
            ! static stability (use virtual potential temperature)
            ! dv now calculated at interface height (lineair in pressure)
            !
            arg=(log(phlb(i,j,l+1))-log(pf(i,j,l)))/(log(pf(i,j,l+1))-log(pf(i,j,l)))
            zthva = zthvk(i,j,l) + arg * (zthvk(i,j,l+1)-zthvk(i,j,l))
            zsstab = grav/zthva*(zthvk(i,j,l+1) - zthvk(i,j,l))/zdz
            !
            ! gradient Richardson number
            !
            zrinub(i,j,l)  = zsstab/zdup2(i,j,l)
            !
            ! stability functions
            !
            zfunst = max(1. - 18.*zrinub(i,j,l),0.)
            zfunst = sqrt(zfunst)
            !
            zfstabh = 1.0 / (1.0 + 10.0*zrinub(i,j,l)*(1.0+8.0*zrinub(i,j,l)))
            !
            ! neutral diffusion coefficient
            !
            zkvn = cml2 * sqrt(zdup2(i,j,l))
            !
            ! correction with stability functions
            !
            jq = jqif( zrinub(i,j,l),0.0 )  ! thus: if (zrinub>0) then jq=1
            wkvf(i,j,l)=jq*(max(ckmin,zkvn*zfstabh))+(1-jq)*(max(ckmin,zkvn*zfunst))
          end do
        end do
      end do
      !
      ! Determine the boundary layer height
      !--------------------------------------------------------------------------------
      ! based on: Vogelezang and Holtslag, 1996, Bound. Layer Meteorol., 81, 245-269.
      !--------------------------------------------------------------------------------

      if (okdebug) write(*,*) '  boundary layer height'

      do j=jsr(region),jer(region)
        do i=isr(region),ier(region)
          ! compute bottom level virtual temperature and heat flux and Monin-Obukhov Length
          wtseff(i,j)  = wthm(i,j,1)*(1. + zcrdq*q(i,j,1))
          wheatv(i,j)  = wheat(i,j) + zcrdq*wthm(i,j,1)*wqflx(i,j)
          wobkl(i,j)   = -wtseff(i,j)*ustr(i,j)**3 / (grav*vkarman*(wheatv(i,j)+sign(1.e-10,wheatv(i,j))) )
          ! initialize
          pblh(i,j)   = 0.0
          jwork  = 1
          wrino  = 0.0
          do l=1,lm(region)

            vvk   = (uwind(i,j,l)-uwind(i,j,1))**2 + (vwind(i,j,l)-vwind(i,j,1))**2 + &
                                     zacb*ustr(i,j)*ustr(i,j)
            vvk   = max(vvk,tiny)

            tkv   = wthm(i,j,l)*(1. + zcrdq*q(i,j,l))
            dtkv  = tkv-wtseff(i,j)
            !
            ! Bulk Richardson number
            !
            zrino = grav*dtkv * (whm(i,j,l)-whm(i,j,1)) / (wtseff(i,j)*vvk)
            zrino = zrino+sign(1.e-10,zrino) ! prevent zrino becoming zero
            zrino = jwork*zrino

            jq    = jqif(ricr,zrino)  ! thus: if (zrino < ricr) then jq=1
            !
            ! calculate pblh from linear interpolation in Ri(bulk)
            !
            if ( l == 1 ) then
              pblh(i,j) = jq*pblh(i,j) + (1-jq) *  &
                    (                (ricr-wrino)/(zrino-wrino)*(whm(i,j,l)             ) )
            else
              pblh(i,j) = jq*pblh(i,j) + (1-jq) *  &
                    ( whm(i,j,l-1) + (ricr-wrino)/(zrino-wrino)*(whm(i,j,l)-whm(i,j,l-1)) )
            end if
            !
            ! first time zrino > ricr we set jwork to zero, thus all further times zrino=0 
            ! and pblh does not change anymore
            !
            jwork = jq*jwork
            !
            ! if pblh is found already avoid dividing by zero next level by a faked value
            ! of wrino (jwork=0 already)
            !
            wrino = zrino+(1-jwork)*0.1

          end do

          pblh(i,j)  = (1-jwork)*pblh(i,j) + whm(i,j,lm(region))*jwork
          !
          ! second calculation of pblh including excess surface temperature using
          ! a convective velocity scale wsc (wm, not equal to w*!!!), using result
          ! of the first calculation of pblh
          !
          jq         = jqif(wheatv(i,j),0.) ! thus: if (wheatv(i,j) > 0.) then jq=1
          jwork = jq
          fmt = ( jq*(1.0 - binm*pblh(i,j)/wobkl(i,j)) + (1.0-jq) )**onet
          wsc     =  ustr(i,j)*fmt
          zexcess =  wheatv(i,j)*fak/wsc
          !
          ! improve pblh estimate under convective conditions using
          ! convective temperature excess (zexcess)
          !
          thvparc = wtseff(i,j)+jq*zexcess
          !
          jwork  = 1
          wrino  = 0.0
          !
          do l=2,lm(region)
            !
            ! Bulk Richardson number:
            !  if stable corrected with extra shear term
            !  if unstable now also corrected for temperature excess
            !
            vvk   = (uwind(i,j,l)-uwind(i,j,1))**2 + (vwind(i,j,l)-vwind(i,j,1))**2 + &
                                      zacb*ustr(i,j)*ustr(i,j)
            vvk   = max(vvk,tiny)
            tkv   = wthm(i,j,l)*(1. + zcrdq*q(i,j,l))
            zrino = grav*(tkv-thvparc) * (whm(i,j,l)-whm(i,j,1))/(vvk*wtseff(i,j))
            zrino = zrino+sign(1.e-10,zrino) ! prevent zrino becoming zero
            zrino = zrino*jwork

            jq    = jqif(ricr,zrino)  ! thus: if (zrino < ricr) then jq=0
            !
            ! calculate pblh from linear interpolation in Ri(bulk)
            !
            pblh(i,j)  = jq*pblh(i,j)+(1-jq)* (whm(i,j,l-1)+(ricr-wrino)/ &
                                                (zrino-wrino)*(whm(i,j,l)-whm(i,j,l-1)))
            !
            ! first time zrino > ricr we set jwork to zero, thus all further times zrino=0 
            ! and pblh does not change anymore
            !
            jwork = jq*jwork
            !
            ! if pblh is found already avoid dividing by zero next level by a faked value
            ! of wrino (jwork=0 already)
            !
            wrino = zrino+(1-jwork)*0.1

          end do

          pblh(i,j) = (1-jwork)*pblh(i,j) + jwork*whm(i,j,lm(region))

        end do
      end do

      pblh = max(pblh, 100.0) ! set minimum value for pblh

      !
      ! Diffusion coefficients in the surface layer
      !
      !--------------------------------------------------------------------------------
      ! the revised LTG scheme (Beljaars and Viterbo, 1998)
      ! Modification - April 1, 1999: prevent zpblk to become too small in very stable conditions
      !--------------------------------------------------------------------------------

      if (okdebug) write(*,*) '  diffusion coeff'

      ! initialize output array with minimum value
      kvhmin = 0.1
      !
      ! Loop to calculate the diffusivity
      !
      do l=1,lm(region)-1
        do j=jsr(region), jer(region)
          do i=isr(region), ier(region)


            jq  = jqif(pblh(i,j),(gph(i,j,l+1)-gph(i,j,1))) ! if top of level hh < pblh jq=1 inside BL
            z   = gph(i,j,l+1)-gph(i,j,1)
            zh  = jq*z/pblh(i,j) ! only defined in BL (jq=1)

            ! zl must not be zero
            zl  = jq*z/wobkl(i,j)+(1-jq) ! z/L outside BL = 1
            zzh = jq*(1.-zh)**2           ! (1-z/h)^2 only in BL

            jq1 = jqif(wheatv(i,j),0.)    ! jq1 is 1 in unstable case
            jq2 = jq*(1-jq1)              ! jq2 is 1 in stable BL
            jq3 = jq*jq1                  ! jq3 is 1 in unstable BL

            jq1 = jqif(sffrac,zh)         ! jq1 is 1 in surface layer
            jq4 = jq3*jq1                 ! jq4 is 1 in unstable surface layer
            jq5 = jq3*(1-jq1)             ! jq5 is 1 outside unstable surface layer
            !
            ! calculate coefficients for momentum, the values for heat are obtained by dividing by Prantl number
            !
            ! stable and neutral:
            !
            cml2    = (1./( 1./(vkarman*z) + 1./450. ))**2.
            jqq     = jqif(zrinub(i,j,l),0.)
            !
            ! stability function for momentum
            !
            zfstabh  = jqq/(1.+10.*zrinub(i,j,l)*sqrt(1.+jqq*zrinub(i,j,l))) + (1-jqq)

            zpblk   = cml2 * sqrt(zdup2(i,j,l))*zfstabh
            zkvh    = jq2*max(zpblk,kvhmin)+(1-jq2)*wkvf(i,j,l) ! take free troposp. values outside BL
            !
            ! unstable case in surface layer
            !
            zslask = 1.-betah*zl
            zslask = jq4*zslask+(1-jq4)
            zpblk  = (1-jq4)*zkvh+jq4*(ustr(i,j)*vkarman*z*zzh*zslask**(1./3.))
            !
            ! unstable case above surface layer
            !
            ! evaluate stability function for momentum at height z = 0.1*pblh (top of surface layer)
            !
            zslask = 1.-ssfrac*betah*pblh(i,j)/wobkl(i,j)
            zslask = jq5*zslask+(1-jq5)
            fmt    = zslask**(1./3.)
            !
            ! use surface layer definition for w_m
            wsc    = ustr(i,j)*fmt
            zpblk  = (1-jq5)*zpblk + jq5*wsc*vkarman*z*zzh
            !
            ! Determine Prandt Number
            !
            ! Pr-number is assumed to be constant throughout the CBL, i.e. in surface and mixed layer
            ! NOTE: this is different from the original formulation
            ! zwstr not used if wheatv<0.
            !
            zwstr  = (abs(wheatv(i,j))*grav*pblh(i,j)/wtseff(i,j))**(1./3.)         ! bh

            obukov = jq3*wobkl(i,j)-(1-jq3)
            term1 = (1.-ssfrac*betah*pblh(i,j)/obukov)**(1./3.)
            term2 = sqrt(1.-ssfrac*betah*pblh(i,j)/obukov)
            term3 = ccon*fakn*zwstr/(ustr(i,j)*(1.-ssfrac*betah*pblh(i,j)/obukov)**(1./3.))

            pr = jq3*(term1/term2+term3)+(1-jq3)
            !
            ! NOTE that kvh applies to the top of the level
            !
            kvh(i,j,l) = jq3*max(zpblk/pr,kvhmin)+(1-jq3)*zkvh

          end do
        end do
      end do

      ! calculate vertical diffusion
      dkg = 0.0
      do  l=1,lmax_conv-1
        do  j=jsr(region), jer(region)
          do  i=isr(region), ier(region)
            dkg(i,j,l)=max(0.,kvh(i,j,l))*2.*(m(i,j,l+1)+m(i,j,l))/(gph(i,j,l+2)-gph(i,j,l))**2
            dkg(i,j,l)=dkg(i,j,l)*czetak
            ! CMK dec 2002 ----> gph changed to 1--->lm+1 boundaries
          enddo
        enddo
      enddo
      dkg(:,:,lmax_conv) = 0.0

      !if(region==1)call dumpfield(1,'dkg_new_1.hdf',im(region),jm(region),lm(region),1,kvh,'kzz')
      !if(region==1)call dumpfield(1,'dkg_new_1.hdf',im(region),jm(region),1,1,pblh,'pblh')
      !if(region==2)call dumpfield(0,'dkg_new_2.hdf',im(region),jm(region),lm(region),1,kvh,'kzz')

      deallocate(wkvf)
      deallocate(zdup2)
      deallocate(zrinub)
      deallocate(wthm)
      deallocate(zthvk)
      deallocate(pf)
      deallocate(whm)
      deallocate(ustr)
      deallocate(wheat)
      deallocate(wqflx)
      deallocate(wobkl)
      deallocate(wheatv)
      deallocate(wtseff)
      deallocate(uwind)
      deallocate(vwind)
!cmk      deallocate(pblh)
      nullify(pblh)
      deallocate(kvh)

      if (okdebug) write(*,*) 'start of bldiff'

    end subroutine bldiff
    

  end subroutine calc_kzz


  ! ***


  !-----------------------------------------------
  !
  ! Purpose:
  ! -------
  ! this subroutine calculate the min,mean,max of a real field
  !
  ! External
  ! --------
  ! none
  !
  ! Reference
  ! ---------
  ! none
  !------------------------------------

  subroutine dd_field_statistics(name,field,ix,jx)

    ! --- in/out -------------------------------
    
    character(len=*),intent(in)        :: name
    integer,intent(in)                 :: ix,jx
    real,dimension(ix,jx),intent(in)   :: field
    
    ! --- local --------------------------------
    
    integer :: i,j,ntel_non_zero
    real :: maxf,minf,meanf,mean_non_zero
    
    ! --- begin -------------------------------
    
    maxf=-1e20
    minf=1e12
    meanf=0.
    mean_non_zero=0.
    ntel_non_zero=0
    do i=1,ix
    do j=1,jx
      meanf=meanf+field(i,j)
      maxf=max(maxf,field(i,j))
      minf=min(minf,field(i,j))
      if (field(i,j).ne.0) then
        ntel_non_zero=ntel_non_zero+1
        mean_non_zero=mean_non_zero+field(i,j)
      endif
    enddo
    enddo
    meanf=meanf/ix/jx
    if (ntel_non_zero.gt.0) mean_non_zero= mean_non_zero/ntel_non_zero
    write(6,'(a10,4(a3,1x,1pe10.3))') name,'max',maxf,'min',minf,'mean',meanf,'mn0',mean_non_zero

  end subroutine dd_field_statistics


  ! ****


  !------------------------------------
  !
  ! Purpose:
  ! -------
  ! this subroutine calculate the min,mean,max of a field of int8
  !
  ! External
  ! --------
  ! none
  !
  ! Reference
  ! ---------
  ! none
  !------------------------------------

  subroutine dd_field_statistics_i8(name,field,ix,jx)

    ! --- in/out ----------------------------

    character(len=*),intent(in)                   :: name
    integer,intent(in)                            :: ix,jx
    integer(kind=1),dimension(ix,jx),intent(in)   :: field

    ! --- local ----------------------------

    integer :: i,j,ntel_non_zero
    real :: maxf,minf,meanf,mean_non_zero
    
    ! --- begin ---------------------------
    
    maxf=-1e20
    minf=1e12
    meanf=0.
    mean_non_zero=0.
    ntel_non_zero=0
    do i=1,ix
      do j=1,jx
        meanf=meanf+field(i,j)
        maxf=max(maxf,real(field(i,j)))
        minf=min(minf,real(field(i,j)))
        if (field(i,j).ne.0) then
          ntel_non_zero=ntel_non_zero+1
          mean_non_zero=mean_non_zero+field(i,j)
        endif
      enddo
    enddo
    meanf=meanf/ix/jx
    if (ntel_non_zero.gt.0) mean_non_zero= mean_non_zero/ntel_non_zero
    write(6,'(a10,4(a3,1x,1pe10.3))') name,'max',maxf,'min',minf,'mean',meanf,'mn0',mean_non_zero

  end subroutine dd_field_statistics_i8
  
  
  ! ***


  !
  ! define vectorizable 'if' function
  ! if xxxz>yyyz jqif=1 else jqif=0
  !

  integer function jqif(xxxz,yyyz)

    ! --- in/out ---------------------------
    
    real, intent(in)    ::  xxxz, yyyz

    ! --- begin ----------------------------
    
    !
    !jqif = nint( 0.5 - sign( 0.5, -dim(xxxz,yyyz)) )
    !
    jqif = nint( 0.5 + sign( 0.5, xxxz-yyyz ) )

  end function jqif


end module diffusion

