!
! open and read all HDF data 
! perform some meteo_dependend calculations
!
module meteo

  use dims, only : nregions

  implicit none
  
  ! --- in/out -------------------------------------

  private

  public  ::  Meteo_Init, Meteo_Done

  public :: new_cloud
  public :: read_oro, read_tempq, read_cloud, read_kzz, read_conv, read_wind

  ! --- const ----------------------------------
  
  ! module name
  character(len=*), parameter  ::  mname = 'module Meteo'
  

  ! --- var -------------------------------------

  ! public variables, types

  logical,dimension(nregions)          :: new_cloud = .false.

  ! --- local -----------------------------------


contains


  ! ==========================================================

  
  subroutine Meteo_Init( status )
  
    use ParTools, only : myid, root
    use MeteoFiles, only : MeteoFiles_Init

    ! --- in/out -------------------------------
        
    integer, intent(out)         ::  status
    
    ! --- const ----------------------------------
    
    character(len=*), parameter  ::  rname = mname//', Meteo_Init'
    
    ! --- begin ----------------------------
    
    ! meteo files could be read by all processors ...
    ! (a file is read by only one pe, but not neccessary the root pe)
    call MeteoFiles_Init( status )
    if (status/=0) then; write (*,'("ERROR in ",a)') rname; status=1; return; end if
    
    ! ok
    status = 0

  end subroutine Meteo_Init


  ! ***


  subroutine Meteo_Done( status )

    use ParTools, only : myid, root
    use MeteoFiles, only : MeteoFiles_Done

    ! --- in/out -------------------------------
        
    integer, intent(out)         ::  status
    
    ! --- const -------------------------------
    
    character(len=*), parameter  ::  rname = mname//', Meteo_Done'
    
    ! --- begin --------------------------------

    ! meteo files could be read by all processors ...
    ! (a file is read by only one pe, but not neccessary the root pe)
    call MeteoFiles_Done( status )
    if (status/=0) then; write (*,'("ERROR in ",a)') rname; status=1; return; end if
    
    ! ok
    status = 0

  end subroutine Meteo_Done



  ! ============================================================
  

  subroutine read_oro

    use dims,        only: im,jm, itau, nlon360, nlat180
    use dims,        only: idate
    use global_data, only: oro, oro1x1
    use TMkit, only : TDate, NewDate, IncrDate, operator(+)
    use TMkit, only : ReadField
    use MeteoFiles, only : tmmd, sourcekey, lli_1x1
    use datetime, only : tau2date
    use toolbox , only : coarsen_emission
    use toolbox , only : escape_tm

    use ParTools,   only : npes, myid, root
#ifdef MPI
    use mpi_const,  only : mpi_comm_world, ierr, my_real
#endif

    ! --- const ----------------------------------
    
    character(len=*), parameter  ::  rname = mname//', Read_Oro'
    
    !integer,parameter     ::  add_field = 0
    integer,parameter     ::  avg_field = 1

    ! --- local ----------------------------------

    type(TDate)        ::  t1, t2
    integer            ::  fcday
    integer            ::  status

    integer    :: region

    integer, dimension(6) :: idater
    !character(len=6)   :: stream   !e.g od/fc/
    !character(len=12)  :: surfdir  ! e.g. 2001/07/sfc/
    !character(len=8)   :: dates

    ! --- begin -----------------------------------
    
!WP!    if (myid==root) then

      if(myid.eq.root) print *, 'read_oro: reading surface altitude'
      call tau2date(itau-21*3600, idater)    ! CHECK TIMING   ;;;
      idater(4) = 21   ! data stored from 21 day1 - 21 day2

      t1 = NewDate( time6=idater )
      t2 = t1 + IncrDate( hour=24 )
      fcday = 0

      call ReadField( tmmd, sourcekey(1), 'oro', t1, t2, fcday, &
                            lli_1x1, 'n', oro1x1, status )
      if (status/=0) call escape_tm('read_oro: ERROR in surface file')

      call coarsen_emission('oro',nlon360,nlat180,oro1x1,oro,avg_field)    

!WP!    end if !WP! alle Pe's from here

    ! distribute to other processores ?
!WP!    if ( npes > 1 ) then
!WP!#ifdef MPI
!WP!      do region = 1, nregions
!WP!         call mpi_bcast( oro(region)%surf, im(region)*jm(region), &
!WP!                                     my_real,root,mpi_comm_world,ierr)
!WP!      end do
!WP!      call mpi_bcast( oro1x1, nlon360*nlat180, &
!WP!                                     my_real,root,mpi_comm_world,ierr)
!WP!#else
!WP!      stop 'please implement for non-mpi parallel library.'
!WP!#endif
!WP!    end if

  end subroutine  read_oro


  ! ***
  

  !-----------------------------------------------------------------------
  !
  !****   rtempq  - input of meteorolgy fields needed for chemistry 
  !
  !       programmed by           aj      KNMI            29-1-1996 
  !       programmed by           mk      IMAU            29-8-2000
  !
  !       purpose
  !       -------
  !       input of temperature and specific humidity  
  !       in hdf per region
  !
  !       interface
  !       ---------
  !       call rtempq(region)
  !       region: region for which the t and q fields are read...
  !           gph is calculated
  !
  !       method
  !       ------
  !       interface : use meteofiles
  !
  !       externals
  !       ---------
  !       subroutines:    read_meteo
  !
  !       reference
  !       ---------       
  !       none
  !----------------------------------------------------------------------

  subroutine read_tempq(region)

    use Dims,        only : cdebug, kdebug, okdebug, kmain
    use Dims,        only : idate, itau, revert, newsrun
    use Dims,        only : im, jm, lm
    use Dims,        only : xbeg, dx, xref, ybeg, dy, yref
    use Dims,        only : at, bt
    use binas,       only : grav
    use global_data, only : meteo_dat, mass_dat, oro, wind_dat
    use datetime,    only : tau2date, tstamp
    use toolbox,     only : escape_tm

    use ParTools,    only : npes, myid, root
#ifdef MPI
    use mpi_const, only : mpi_comm_world, ierr, my_real
#endif

    use TMkit, only : TDate, NewDate, IncrDate, operator(+)
    use TMkit, only : ReadField
    use TMkit, only : ReadUVSP
    use MeteoFiles, only : tmmd, sourcekey, lli, levi

    ! in/out
    integer,intent(in)                   :: region

    ! --- const ----------------------------------
    
    character(len=*), parameter  ::  rname = mname//', read_tempq'
    
    ! --- local ----------------------------------
    
    real,dimension(:,:,:),pointer        :: t,q,pu,pv
    real,dimension(:,:,:),pointer        :: gph
    real,dimension(:,:),pointer          :: ps

    real, dimension(:,:), allocatable    :: p_beg, p_mid, p_end  !mkadj

    integer                              :: i,j,l
    real                                 :: tv,pdown,pup
    integer,dimension(6)                 :: idater
    integer :: nsend

    type(TDate)        ::  t1, t2
    integer            ::  fcday
    integer            ::  status

    ! --- begin -----------------------------

    ps => mass_dat(region)%p
    gph => meteo_dat(region)%gph
    t => meteo_dat(region)%t
    q => meteo_dat(region)%q

!WP!    if (myid==root) then ! only root does IO

       if ( cdebug.and.myid==root )  call tstamp(kdebug,itau,'rtempq')
       if ( okdebug.and.myid==root )  call tstamp(kmain,itau,'rtempq')

       if ( revert == 1 )  then
          call tau2date(itau,idater)   
       else 
          !get the temper. 6 hours before....e.g. 21-->03
          call tau2date(itau-3600*6,idater)
       end if

       ! first time might be corresponding to time in file (0, 6, 12, 18)
       if ( newsrun ) then
          !mk check for revert run!
          if (mod(idate(4),6) /= 3) call tau2date(itau-3*3600,idater)
       end if

       ! set times to t1+[0,6]
       t1 = NewDate( time6=idater )
       t2 = t1 + IncrDate( hour=6 )
       fcday = 0
       
       call ReadField( tmmd, sourcekey(region), 'T', t1, t2, fcday, &
                             lli(region), 'n', levi, &
                             ps(1:im(region),1:jm(region)), &
                             t, status )
       if (status/=0) call escape_tm( 'read_tempq: ERROR reading temperature')

       call ReadField( tmmd, sourcekey(region), 'Q', t1, t2, fcday, &
                             lli(region), 'n', levi, &
                             ps(1:im(region),1:jm(region)), &
                             q, status )
       if (status/=0) call escape_tm( 'read_tempq: ERROR reading humidity')

       if ( newsrun )  &
            call tstamp(kmain,itau,'read_tempq: read temperature and humidity')
       if ( cdebug ) &
            call tstamp(kdebug,itau,'read_tempq: read temperature and humidity')
       if ( okdebug ) &
            call tstamp(kmain,itau,'read_tempq: read temperature and humidity')
       if (revert == 1) then 
          gph(:,:,1)=oro(region)%surf/grav   !oro is stored in g*mif (revert == 1) then 
          do l=1,lm(region)-1
             do j=1,jm(region)
                do i=1,im(region)
                   tv = t(i,j,l)*(1. + 0.608*q(i,j,l))
                   pdown = at(l) + bt(l)*ps(i,j)
                   pup   = at(l+1) + bt(l+1)*ps(i,j)
                   gph(i,j,l+1) = gph(i,j,l) + tv*287.05*alog(pdown/pup)/grav   !rgas in diffrent units!
                   ! note dec 2002 (MK) gph now from 1--->lm+1
                enddo
             enddo
          enddo
          gph(:,:,lm(region)+1) = 200000.0   !set top of atmosphere at 200 km !note_mk: now in lm+1
       else   ! revert == -1
          nullify(pu)
          nullify(pv)
          pu => wind_dat(region)%pu_t
          pv => wind_dat(region)%pv_t
          allocate( p_beg(size(ps,1),size(ps,2)) )
          allocate( p_mid(size(ps,1),size(ps,2)) )
          allocate( p_end(size(ps,1),size(ps,2)) )
          call ReadUVSP( tmmd, sourcekey(region), t1, t2, fcday, &
                      lli(region), levi, &
                      p_beg(1:im(region),1:jm(region)), &
                      p_mid(1:im(region),1:jm(region)), &
                      p_end(1:im(region),1:jm(region)), &
                      pu(0:im(region),1:jm(region)  ,1:lm(region)), &
                      pv(1:im(region),1:jm(region)+1,1:lm(region)), &
                      status )
          if (status/=0) call escape_tm( 'read_uvsp: ERROR reading uvsp')
          !mkadj: use the BEGIN pressure to calculate gph
          gph(:,:,1)=oro(region)%surf/grav   !oro is stored in g*mif (revert == 1) then 
          do l=1,lm(region)-1
             do j=1,jm(region)
                do i=1,im(region)
                   tv = t(i,j,l)*(1. + 0.608*q(i,j,l))
                   pdown = at(l) + bt(l)*p_beg(i,j)
                   pup   = at(l+1) + bt(l+1)*p_beg(i,j)
                   gph(i,j,l+1) = gph(i,j,l) + tv*287.05*alog(pdown/pup)/grav   !rgas in diffrent units!
                   ! note dec 2002 (MK) gph now from 1--->lm+1
                enddo
             enddo
          enddo
          gph(:,:,lm(region)+1) = 200000.0   !set top of atmosphere at 200 km !note_mk: now in lm+1
          deallocate( p_beg )
          deallocate( p_mid )
          deallocate( p_end )
          nullify(pu)
          nullify(pv)
       endif
!WP!    end if !all PE's from here

!WP!    if ( npes > 1 ) then
!WP!#ifdef MPI
!WP!      nsend=(im(region)+4)*(jm(region)+4)
!WP!      call mpi_bcast(ps,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=im(region)*jm(region)*(lm(region)+1)
!WP!      call mpi_bcast(gph,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=im(region)*jm(region)*lm(region)
!WP!      call mpi_bcast(t,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=im(region)*jm(region)*lm(region)
!WP!      call mpi_bcast(q,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      if ( myid == root ) print *,'read_tempq: ps,t,q,gph data broadcasted '
!WP!#else
!WP!      stop 'please implement for non-mpi parallel library.'
!WP!#endif
!WP!    end if

    ! done
    nullify( ps )
    nullify( gph )
    nullify( t )
    nullify( q )

  end subroutine read_tempq



   !-----------------------------------------------------------------------
   !
   !****   read_cloud - input of meteorolgy fields needed for chemistry 
   !
   !       programmed by           aj      KNMI            29-1-1996 
   !       programmed by           mk      IMAU            29-8-2000
   !
   !       purpose
   !       -------
   !       input of lwc, iwc, and cc
   !       in hdf per region
   !
   !       interface
   !       ---------
   !       call read_cloud(region)
   !       region: region for which the lwc,iwc,cc are read

   !       method
   !       ------
   !       use meteofiles
   !
   !       externals
   !       ---------
   !       subroutine: read_meteo(region)
   !
   !       reference
   !       ---------       
   !       none
   !----------------------------------------------------------------------

  subroutine read_cloud(region)

    use dims,        only : cdebug, kdebug, okdebug, kmain
    use dims,        only : itau, idate, revert, newsrun
    use Dims,        only : im, jm, lm
    use Dims,        only : xbeg, dx, xref, ybeg, dy, yref
    use global_data, only : cloud_dat, meteo_dat, mass_dat
    use datetime,    only : tau2date, tstamp
    use toolbox,     only : escape_tm

    use ParTools,    only : npes, myid, root
#ifdef MPI
    use mpi_const, only : mpi_comm_world,ierr,my_real
#endif

    use TMkit, only : TDate, NewDate, IncrDate, operator(+)
    use TMkit, only : ReadField
    use MeteoFiles, only : tmmd, sourcekey, lli, levi

    ! --- in/out ---------------------------------------
    
    integer,intent(in) :: region

    ! --- local ----------------------------------------
    
    real,dimension(:,:,:),pointer        :: cc,lwc,iwc,ccu, cco
    real,dimension(:,:),pointer          :: ps
    integer,dimension(6)                 :: idater
    integer                              :: nsend

    type(TDate)        ::  t1, t2
    integer            ::  fcday
    integer            ::  status

    ! --- begin ---------------------------------------

    if ( cdebug )  call tstamp(kdebug,itau,'read_cloud')
    if ( okdebug )  call tstamp(kmain,itau,'read_cloud')

    ps => mass_dat(region)%p
    lwc=> cloud_dat(region)%lwc 
    iwc=> cloud_dat(region)%iwc 
    cc=>  cloud_dat(region)%cc
    ccu=>  cloud_dat(region)%ccu
    cco=>  cloud_dat(region)%cco

!    if (myid==root) then ! WP only root does IO

       if (revert == 1)  then
          call tau2date(itau,idater)   
       else 
          !get the temper. 6 hours before....e.g. 21-->03
          call tau2date(itau-3600*6,idater)
       end if

       ! first time might be corresponding to time in file (0, 6, 12, 18)
       if ( newsrun ) then
          !mk check for revert run!
          if (mod(idate(4),6) /= 3) call tau2date(itau-3*3600,idater)
       end if

       ! set times to t1+[0,6]
       t1 = NewDate( time6=idater )
       t2 = t1 + IncrDate( hour=6 )
       fcday = 0
       
!       call ReadField( tmmd, sourcekey(region), 'CC', t1, t2, fcday, &
!                             lli(region), 'n', levi, &
!                             ps(1:im(region),1:jm(region)), &
!                             cc, status )
!       if (status/=0) call escape_tm('read_cloud: ERROR reading cloud cover')

!      call ReadField( tmmd, sourcekey(region), 'CCU', t1, t2, fcday, &
!                            lli(region), 'n', levi, &
!                            ps(1:im(region),1:jm(region)), &
!                            ccu, status )
!      if (status/=0) call escape_tm('read_cloud: ERROR reading underfeet cloud cover')

!        call ReadField( tmmd, sourcekey(region), 'CCO', t1, t2, fcday, &
!                            lli(region), 'n', levi, &
!                            ps(1:im(region),1:jm(region)), &
!                            cco, status )
!       if (status/=0) call escape_tm('read_cloud: ERROR reading overhead cloud cover')

!      call ReadField( tmmd, sourcekey(region), 'CLWC', t1, t2, fcday, &
!                            lli(region), 'n', levi, &
!                            ps(1:im(region),1:jm(region)), &
!                            lwc, status )
!      if (status/=0) call escape_tm('read_cloud: ERROR reading liquid water content')

!      call ReadField( tmmd, sourcekey(region), 'CIWC', t1, t2, fcday, &
!                            lli(region), 'n', levi, &
!                            ps(1:im(region),1:jm(region)), &
!                            iwc, status )
!      if (status/=0) call escape_tm('read_cloud: ERROR reading ice water content')

!      lwc=max(0.,lwc)
!      iwc=max(0.,iwc)
       cc=min(1.0,max(0.,cc))   !CMK could be slightly larger than 1!

!    end if !WP! all processors from here

!    if ( npes > 1 ) then
!#ifdef MPI
!      nsend=im(region)*jm(region)*lm(region)
!      call mpi_bcast(lwc,nsend,my_real,root,mpi_comm_world,ierr)
!      call mpi_bcast(iwc,nsend,my_real,root,mpi_comm_world,ierr)
!      call mpi_bcast( cc,nsend,my_real,root,mpi_comm_world,ierr)
!      if ( myid == root ) print*,'read_cloud: cloud data broadcasted '
!#else
!      stop 'please implement for non-mpi parallel library.'
!#endif
!    end if

    new_cloud(region) = .true.   !set flag for photolysis 

    if ( myid == root ) then
      if ( okdebug ) call tstamp(kmain,itau,'read_cloud:  lwc, iwc, and cc')
    end if

    ! done
    nullify(lwc)
    nullify(iwc)
    nullify(cc)
    nullify(ccu)
    nullify(cco)

  end subroutine read_cloud


  !
  ! read kzz file
  !

  subroutine read_kzz(region)
  
    integer, intent(in)  ::  region

    write (*,'("ERROR - Sorry, kzz files are not supported anymore.")')
    write (*,'("ERROR - Please use online diffusion.")')

!    use dims
!    use global_data, only: meteo_dat, conv_dat, mass_dat
!    use datetime,    only : tau2date, tstamp
!    use toolbox,     only : escape_tm
!#ifdef MPI
!    use mpi_const,only : myid,root,mpi_comm_world,ierr,my_real
!#endif
!
!    implicit none
!
!    ! in/out
!    integer,intent(in) :: region
!
!    ! local
!    real,dimension(:,:,:),pointer        :: m       !air mass (kg)
!    real,dimension(:,:,:),pointer        :: dkg     !vert diff (ks/s)
!    real,dimension(:,:,:),pointer        :: gph     !geopotential (m)
!    real,dimension(:,:,:),allocatable    :: ds
!
!    integer,dimension(6)      :: idater
!    integer                   :: i,j,l
!    integer                   :: nsend
!
!    ! start
!
!    if ( cdebug )   call tstamp(kdebug,itau,'read_kzz')
!    if ( okdebug )  call tstamp(kmain,itau,'read_kzz')
!
!    m => mass_dat(region)%m_t
!    gph => meteo_dat(region)%gph
!    dkg => conv_dat(region)%dkg
!    ! note kzz comes 3-hourly!    
!
!#ifdef MPI
!    if (myid==root) then ! WP only root does IO
!#endif
!
!       select case(revert)
!       case(1)
!          ! data are from 21-00, 00-03, etc
!          call  tau2date(itau,idater) 
!       case(-1)
!          ! in reverse get fields 3 hours before!
!          call  tau2date(itau-3*3600,idater) 
!       case default
!       end select
!
!       fkzz(region)%field = 'kz'
!       fkzz(region)%idate = idater
!       fkzz(region)%lon_first = real(xbeg(region)) + 0.5*dx/xref(region)
!       fkzz(region)%lat_first = real(ybeg(region)) + 0.5*dy/yref(region)
!       fkzz(region)%lon_inc   = dx/xref(region)
!       fkzz(region)%lat_inc   = dy/yref(region)
!       fkzz(region)%lon_n     = im(region) 
!       fkzz(region)%lat_n     = jm(region)
!       fkzz(region)%lm        = lm(region)
!       allocate(ds(im(region),jm(region),lm(region)))
!
!       if ( fkzz(region)%flag == 0 ) then 
!          call read_meteo(region, fkzz(region), ds, im(region), &
!               jm(region), lm(region))
!       else 
!          call escape_tm( 'read_kzz: Flag in kzz not OK')
!       end if
!
!       print*,'read_kzz: ifail KZZ ',fkzz(region)%flag,'region ',region, idate
!       if ( fkzz(region)%flag /= 0 ) &
!            call escape_tm('read_kzz: Problem with KZZ file')
!
!       dkg(1:im(region),1:jm(region),1:lmax_conv) = &
!            ds(1:im(region), 1:jm(region), 1:lmax_conv)
!
!       deallocate(ds)
!
!       !               scale diffusion coefficient by czetak
!       !               make diffusion coefficient explicitely non-negative
!       !mk             diffusion is given in m2/s
!       !mk             multiply by rho (rho = m/(area*dz)) gives kg/(m s)
!       !mk             multiply by area gives kg m/s
!       !mk             divide by dz gives kg/s
!       !mk             in convma divided by m and multiplied by dt....
!       !mk             dz**2 = 0.25*(ghp(l+1)-gph(l-1))**2
!       !mk             m     = 0.5*(m(l+1)+m(l))
!       !               dkg(is,l) is the diffusion coefficient at the upper
!       !               border of layer l; so dkg(is,lm)=0.
!
!       do  l=1,lmax_conv-1
!          do  j=1,jm(region)
!             do  i=1,im(region)
!                dkg(i,j,l) = max(0.,dkg(i,j,l))*czetak*2.* &
!                     (m(i,j,l+1)+m(i,j,l))/(gph(i,j,l+2)-gph(i,j,l))**2
!                ! CMK dec 2002 ----> gph changed to 1--->lm+1 boundaries
!             end do
!          end do
!       end do
!       dkg(:,:,lmax_conv) = 0.0
!       if ( okdebug ) call tstamp(kmain,itau,'read_kzz: read and converted kzz')
!
!#ifdef MPI
!    end if !WP! all processors from here
!
!    nsend=im(region)*jm(region)*lmax_conv
!    call mpi_bcast(dkg,nsend,my_real,root,mpi_comm_world,ierr)
!
!    if (myid == root ) print*,'read_kzz: data broadcasted '
!#endif
!
!    nullify(m)
!    nullify(dkg)
!    nullify(gph)

  end subroutine read_kzz



  !-----------------------------------------------------------------------
  !
  !****   read_conv           - read convection info          v 9.1
  !
  !       programmed by           mh      mpi HH          1-oct-1991
  !       modified                                      11-jun-1994
  !       for zoom version MK, Imau    4/2000
  !
  !       purpose
  !       -------
  !       read convection info (entrainment, detrainment rates)
  !
  !       interface
  !       ---------
  !       call read_conv
  !
  !       method
  !       ------
  ! 
  !       read entrainement en detrainement rates for hdf file, 
  !       scale them by czeta.
  !
  !       externals
  !       ---------
  !       subroutines:    read_meteo
  !                       exitus
  !                       tstamp
  !       reference
  !       ---------       
  !       see manual(..)
  !-----------------------------------------------------------------------

  subroutine read_conv(region)

    use dims,        only : cdebug, kdebug, okdebug, kmain
    use Dims,        only : itau, idate, revert, newsrun
    use Dims,        only : im, jm, lm, lmax_conv
    use Dims,        only : xbeg, dx, xref, ybeg, dy, yref
    use Dims,        only : czeta
    use global_data, only : conv_dat, region_dat
    use datetime,    only : tau2date, tstamp
    use toolbox,     only : escape_tm

    use ParTools,    only : npes, myid, root
#ifdef MPI
    use mpi_const, only : mpi_comm_world,ierr,my_real
#endif

    use global_data, only : mass_dat
    use TMPhys, only : ConvCloudDim
    use TMkit, only : TDate, NewDate, IncrDate, operator(+)
    use TMkit, only : ReadField
    use MeteoFiles, only : tmmd, sourcekey, lli, levi

    ! --- in/out -------------------------------------
    
    integer, intent(in)                   :: region

    ! --- local -------------------------------------
    
    type(TDate)        ::  t1, t2
    integer            ::  fcday
    integer            ::  status
    real,dimension(:,:),pointer          :: ps

    real,dimension(:,:,:),allocatable    :: ds

    real,dimension(:,:,:) ,pointer        :: entu,detu,entd,detd
    integer,dimension(:,:), pointer       :: cloud_base, cloud_top, cloud_lfs
    real,dimension(:),pointer             :: dxyp 
    integer,dimension(6)      :: idater
    integer                   :: i,j,l
    real                      :: tote,totd,maxe
    integer                   :: lsave
    integer                   :: nsend
    
    ! --- begin -----------------------------------

    if ( cdebug )   call tstamp(kdebug,itau,'read_conv')
    if ( okdebug )  call tstamp(kmain,itau,'read_conv')

    dxyp => region_dat(region)%dxyp
    entu => conv_dat(region)%entu
    entd => conv_dat(region)%entd
    detu => conv_dat(region)%detu
    detd => conv_dat(region)%detd
    cloud_top => conv_dat(region)%cloud_top
    cloud_base => conv_dat(region)%cloud_base
    cloud_lfs => conv_dat(region)%cloud_lfs

!WP!    if (myid==root) then !only root does IO

      if ( revert == 1 )  then
         call tau2date(itau,idater)   
      else 
         ! get the temper. 6 hours before....e.g. 21-->03
         call tau2date(itau-3600*6,idater)
      end if

      ! first time might be corresponding to time in file (0, 6, 12, 18)
      if ( newsrun ) then
         !mk ALSO for revert run!
         if (mod(idate(4),6) /= 3)  call tau2date(itau-3*3600,idater)
      end if

      ! surface pressure
      ps => mass_dat(region)%p

      ! set times to t1+[0,6]
      t1 = NewDate( time6=idater )
      t2 = t1 + IncrDate( hour=6 )
      fcday = 0

      allocate(ds(im(region),jm(region),lm(region)))

      ! ~~

      call ReadField( tmmd, sourcekey(region), 'eu', t1, t2, fcday, &
                            lli(region), 'n', levi, &
                            ps(1:im(region),1:jm(region)), &
                            ds, status )
      if (status/=0) call escape_tm('read_conv: ERROR reading entr up')

      entu(1:im(region),1:jm(region),1:lmax_conv) = &
           ds(1:im(region), 1:jm(region), 1:lmax_conv)

      ! ~~

      call ReadField( tmmd, sourcekey(region), 'ed', t1, t2, fcday, &
                            lli(region), 'n', levi, &
                            ps(1:im(region),1:jm(region)), &
                            ds, status )
      if (status/=0) call escape_tm('read_conv: ERROR reading entr down')

      entd(1:im(region),1:jm(region),1:lmax_conv) = &
           ds(1:im(region), 1:jm(region), 1:lmax_conv)

      ! ~~

      call ReadField( tmmd, sourcekey(region), 'du', t1, t2, fcday, &
                            lli(region), 'n', levi, &
                            ps(1:im(region),1:jm(region)), &
                            ds, status )
      if (status/=0) call escape_tm('read_conv: ERROR reading detr up')

      detu(1:im(region),1:jm(region),1:lmax_conv) = &
           ds(1:im(region), 1:jm(region), 1:lmax_conv)

      ! ~~

      call ReadField( tmmd, sourcekey(region), 'dd', t1, t2, fcday, &
                            lli(region), 'n', levi, &
                            ps(1:im(region),1:jm(region)), &
                            ds, status )
      if (status/=0) call escape_tm('read_conv: ERROR reading detr down')

      detd(1:im(region),1:jm(region),1:lmax_conv) = &
           ds(1:im(region), 1:jm(region), 1:lmax_conv)

      ! ~~

      deallocate(ds)

      ! ~~
      
      ! Trap zero downdraft fields.
      ! From 2004-05-09 to 2005-07-12, subgrid files od-fc-ml60-glb3x2
      ! have been produced with downdraft entr/detr (entd,detd) equal to zero,
      ! caused by unreproducable compiler failure at ecgate.
      ! Arjo Segers, 2005-07-13 .
      
      if ( any( (/maxval(entu),maxval(detu),maxval(entd),maxval(detd)/) == 0.0 ) ) then
        write (*,'("ERROR - obscure data ranges for entrainments/detrainments:")')
        write (*,'("ERROR -   updraft   entrement : ",2e12.4)') minval(entu), maxval(entu)
        write (*,'("ERROR -   updraft   detrement : ",2e12.4)') minval(detu), maxval(detu)
        write (*,'("ERROR -   downdraft entrement : ",2e12.4)') minval(entd), maxval(entd)
        write (*,'("ERROR -   downdraft detrement : ",2e12.4)') minval(detd), maxval(detd)
        !call escape_tm('ERROR in meteo/read_conv')
      end if

      ! ~~

      do j = 1, jm(region)
        do i = 1, im(region)
          call ConvCloudDim( detu(i,j,1:lmax_conv), entd(i,j,1:lmax_conv), 'u', &
                             cloud_base(i,j), cloud_top(i,j), cloud_lfs(i,j), &
                             status )
          if (status/=0) call escape_tm('read_conv: ERROR computing conv cloud dims')
        end do
      end do
       
       !cmk calculate the rates in kg/gridbox and scale with czeta

       do j=1,jm(region)
          do i=1,im(region)
             entu(i,j,1:lmax_conv) = entu(i,j,1:lmax_conv)*dxyp(j)*czeta
             detu(i,j,1:lmax_conv) = detu(i,j,1:lmax_conv)*dxyp(j)*czeta
             tote = sum(entu(i,j,1:lmax_conv))
             totd = sum(detu(i,j,1:lmax_conv))
             maxe = entu(i,j,1)  ! changed: reported by PB feb 2003
             lsave = 1
             do l=2,lmax_conv
                if (entu(i,j,l) > maxe) then
                   maxe = entu(i,j,l)
                   lsave = l
                end if
             end do
             entu(i,j,lsave) = entu(i,j,lsave) - tote + totd
          end do
       end do
       do j=1,jm(region)
          do i=1,im(region)
             entd(i,j,:) = entd(i,j,:)*dxyp(j)*czeta
             detd(i,j,:) = detd(i,j,:)*dxyp(j)*czeta
             tote = sum(entd(i,j,:))   !total entrainement
             totd = sum(detd(i,j,:))   !total detrainement
             maxe = 0.0
             lsave = lmax_conv
             do l=1,lmax_conv
                if (entd(i,j,l) > maxe) then
                   maxe = entd(i,j,l)
                   lsave = l
                end if
             end do
             entd(i,j,lsave) = entd(i,j,lsave) - tote + totd
          end do
       end do

       if ( okdebug ) &
            call tstamp(kmain,itau, &
            'read_conv: read and corrected convective fluxes')

!WP!    end if !WP! all processors from here
!WP!
!WP!    ! from root to other processors ...
!WP!    if ( npes > 1 ) then
!WP!#ifdef MPI
!WP!      nsend=jm(region)
!WP!      call mpi_bcast(dxyp,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=im(region)*jm(region)*lmax_conv
!WP!      call mpi_bcast(entu,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      call mpi_bcast(entd,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      call mpi_bcast(detu,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      call mpi_bcast(detd,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=im(region)*jm(region)
!WP!      call mpi_bcast(cloud_base,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      call mpi_bcast(cloud_top,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      call mpi_bcast(cloud_lfs,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      if ( myid == root ) print*,'read_conv: convection data broadcasted '
!WP!#else
!WP!      stop 'please implement for non-mpi parallel library.'
!WP!#endif
!WP!    end if

    nullify(dxyp)
    nullify(entu)
    nullify(entd)
    nullify(detu)
    nullify(detd)
    nullify(cloud_top)
    nullify(cloud_base)
    nullify(cloud_lfs)

  end subroutine read_conv
  


  !-----------------------------------------------------------------------
  !
  !****   read_wind           - input of massfluxfields       v 9.1
  !
  !       programmed by           mh      mpi HH          1-oct-1991
  !       modified                                        26-sep-1992
  !       modified                                        23-dec-2002 MK
  !
  !       purpose
  !       -------
  !       input of massfluxfields and surface pressure 
  !
  !       interface
  !       ---------
  !       call read_wind
  !
  !       method
  !       ------
  !       none
  !
  !       externals
  !       ---------
  !       subroutines:    exitus
  !                       dynam0
  !                       tstamp
  !                       geomtryv
  !
  !       reference
  !       ---------       
  !       none
  !-----------------------------------------------------------------------
  !       modified to test the date
  !                               mr, lmce/cfr, 1-oct-1994
  !-----------------------------------------------------------------------
  !       modified according to v 9.1 and added vertical geometry calc.
  !                               aj, KNMI, 30-08-1995
  !----------------------------------------------------------------------
  !       for zoom version: read per region and use io_hdf
  !                               Maarten Krol, Oct, 2002

  subroutine read_wind(region)

    use dims,        only : cdebug, kdebug, kmain
    use Dims,        only : itau, idate, revert, newsrun
    use Dims,        only : im, jm, lm
    use Dims,        only : xbeg, dx, xref, ybeg, dy, yref
    use Dims,        only : xcyc
    use global_data, only : wind_dat, mass_dat
    use redgridZoom, only : calc_pdiff
    use geometry,    only : geomtryv
    use io_hdf,      only : io_write2d_32d, dfacc_create
    use datetime,    only : tau2date, tstamp
    use toolbox,     only : escape_tm

    use TMkit, only : TDate, NewDate, IncrDate, operator(+)
    use TMkit, only : ReadUVSP
    use MeteoFiles, only : tmmd, sourcekey, lli, levi

    use ParTools,    only : npes, myid, root
#ifdef MPI
    use mpi_comm, only : barrier
    use mpi_const, only : mpi_max, mpi_comm_world, my_real, ierr
    use mpi_const, only : ntracetloc
#endif

    ! --- in/out ----------------------------------------
    
    integer,intent(in)  :: region
    
    ! --- local -----------------------------------------

    ! MPI
    integer  :: nsend
    real     :: pdiffmax

    type(TDate)             ::  t1, t2
    integer                 ::  fcday
    integer                 ::  status
    real, allocatable       ::  p_mid(:,:)
    real, allocatable       ::  p_end(:,:)
    real, allocatable,target ::  p_beg(:,:)

    real,dimension(-1:im(region)+2,-1:jm(region)+2) :: pold
    real,dimension(:,:,:),pointer        :: pu,pv,m    !fluxes, air mass(kg)
    real,dimension(:,:),pointer          :: p
    ! pold should have same dimension as p

    real                         :: pdiffmaxl
    integer,dimension(6)         :: idater,idatep
    integer                      :: i,j,l

    integer                      :: io
    integer, external            :: sfstart, sfend
    
    ! --- begin -------------------------------------

    if ( cdebug )  call tstamp(kdebug,itau,'read_wind')

    p => mass_dat(region)%p
    pu => wind_dat(region)%pu_t
    pv => wind_dat(region)%pv_t
    m  => mass_dat(region)%m_t

    select case(revert)
    case ( -1 )
       call tau2date(itau-3600*6, idater)   !backward run: winds -6 hours
       ! for pressure use actual time, end pressure is stored in file!
       idatep = idate
    case ( 1)
       idater = idate
       idatep = idate
    case default
    end select

    ! this holds for revert=1/-1:
    ! first time might not be corresponding to time in file (21,3,9,15)
    if ( newsrun ) then
       if ( mod(idate(4),6) /= 3 ) call tau2date(itau-3600*3,idater)
    else
       pold = p
    end if

!WP!    if (myid==root) then !WP! only root reads

       ! set times to t1+[0,6]
       t1 = NewDate( time6=idater )
       t2 = t1 + IncrDate( hour=6 )
       fcday = 0
       
       allocate( p_mid(size(p,1),size(p,2)) )
       allocate( p_end(size(p,1),size(p,2)) )

       if(revert==-1 .and. newsrun) then  
          ! end pressure is restored already!
          nullify(p)
          allocate( p_beg(size(p,1),size(p,2)) )
          ! let p point to dummy array, !WP! caution, only root now points to
          ! another array, not other PE's!!! 
          p => p_beg  
       endif

       call ReadUVSP( tmmd, sourcekey(region), t1, t2, fcday, &
                      lli(region), levi, &
                          p(1:im(region),1:jm(region)), &
                      p_mid(1:im(region),1:jm(region)), &
                      p_end(1:im(region),1:jm(region)), &
                      pu(0:im(region),1:jm(region)  ,1:lm(region)), &
                      pv(1:im(region),1:jm(region)+1,1:lm(region)), &
                      status )
       if (status/=0) call escape_tm( 'read_uvsp: ERROR reading uvsp')

       !mkadj: get the END pressure for the adjoint
       if (revert == -1) then
         if(newsrun) then  ! when newsrun we want to reset p to old mass_dat, do
!                            this before bcast so that correct (old) p is on all
!                            PE's ! deallocate p_beg now as it is no longer
!                            needed
            nullify(p)
            deallocate( p_beg )
            p=>mass_dat(region)%p  ! all PEs point to mass_dat again
         else           ! if not newsrun, put p_end in p (=>mass_dat on all PEs) and bcast
            p(1:im(region),1:jm(region)) = p_end(1:im(region),1:jm(region))
         endif
       endif

       deallocate( p_mid )
       deallocate( p_end )

       if ( newsrun ) call tstamp(kmain,itau,'read_wind: read massfluxes')
       if ( cdebug ) call tstamp(kdebug,itau,'read_wind: read massfluxes')

!WP!    end if !WP! all processors from here

!WP! In case of revert==1 we read the pressure on root and bcast to all using
!WP! mass_dat%p as the array

!WP! In case of revert==-1 and newsrun==F we read the pressure on root, then replace it with p_end,
!WP! and bcast to all using mass_dat%p as the target array

!WP! In case of revert==-1 and newsrun==T we read the pressure on root into p_beg, discard it,
!WP! replace it with old mass_dat, and bcast to all using mass_dat%p as the target array

!WP!    ! send to other processors ?
!WP!    if ( npes > 1 ) then
!WP!#ifdef MPI
!WP!      nsend=(im(region)+4)*(jm(region)+4)
!WP!      call mpi_bcast(p,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=(im(region)+4)*(jm(region)+4)*lm(region)
!WP!      call mpi_bcast(m,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=(im(region)+2)*(jm(region)+2)*(lm(region)+1)
!WP!      call mpi_bcast(pu,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      nsend=(im(region)+2)*(jm(region)+2)*(lm(region)+1)
!WP!      call mpi_bcast(pv,nsend,my_real,root,mpi_comm_world,ierr)
!WP!      if ( myid == root ) print*,'read_wind: wind data broadcasted '
!WP!#else
!WP!      stop 'please implement for non-mpi parallel library.'
!WP!#endif
!WP!    end if

    ! calculate vertical geomtry and
    ! amount of air transferred in each model step

    if ( newsrun ) then
       !WP! no longer needed, the pointer p already points to the old mass_dat on all PEs and
       !WP! p_beg has already been deallocated on root

       !if (revert == -1) then  ! set pointer to restored p
       !   nullify(p)
       !   if(myid==root) deallocate( p_beg )
       !   p => mass_dat(region)%p
       !endif

       call geomtryv(region)       
    else
       call calc_pdiff(region,p,pold,pdiffmaxl)
#ifdef MPI  
       if (ntracetloc==0) pdiffmaxl=0.0
       call mpi_allreduce(pdiffmaxl,pdiffmax,1,my_real, &
            mpi_max,mpi_comm_world,ierr) 
#else
          pdiffmax = pdiffmaxl
#endif
       if ( myid == root ) then
          print*,'read_wind: Max difference between calculated and read-in ', &
               'pressure (Pa)',pdiffmax
          if ( pdiffmax > 100 ) then
             io = sfstart('pressure.hdf',DFACC_CREATE)
             call io_write2d_32d(io,im(region),'LON',jm(region),'LAT', &
                  p(1:im(region),1:jm(region)),'p',idate)
             call io_write2d_32d(io,im(region),'LON',jm(region),'LAT', &
                  pold(1:im(region),1:jm(region)),'pold',idate)
             print *,sfend(io)
             PRINT *, '====================================================='
             PRINT *, 'read_wind: problem with read-in pressure'
             PRINT *, '====================================================='
             p(:,:) = pold(:,:)

          end if !pdiffmax
       end if !root
    end if !newsrun

    if (xcyc(region)==1) then   !periodic boundary
       m(0           ,:,:) = m(im(region),:,:)
       m(im(region)+1,:,:) = m(         1,:,:)
       pu(0           ,:,:) = pu(im(region),:,:)
       pu(im(region)+1,:,:) = pu(         1,:,:)
    end if

    nullify(p)
    nullify(pu)
    nullify(pv)
    nullify(m)

  end subroutine read_wind


end module meteo
